#include "cbgps.h"
#include "cbgpx.h"
#include <stdio.h>
#include <math.h>
#include <string.h>

/** \file cbcons.c
  * \brief Console application for retrieving GPS positions from a serial port
  */

/** \brief Context structure for the gps callback.
  */
struct _gpsctx {
  /** \brief Handle of the GPX file */
  t_cbgpx  hgpx;
  /** \brief Previous value of the latitude. */
  double   oldlat;
  /** \brief Previous value of the longitude. */
  double   oldlon;
  /** \brief Previous value of the altitude. */
  double   oldalt;
  /** \brief Previous value of UTM easting. */
  double   oldutme;
  /** \brief Previous value of UTM northing. */
  double   oldutmn;
}; /* struct _gpsctx */

/** \brief Callback function retrieving serial port info.
  * Prints information about available ports, baudrates, databits,
  * stopbits etc. to stdout.
  * \param aText the retrieved information text, eg. port name,
  *              name of the available baudrate, number of
  *              databits, stopbits or the name of the parity.
  * \param aType the type of the information in aText.
  * \param aCtx  ignored.
  * \sa cbserinfo()
  */
static void _cbsicb(const char *aText, CBSERINFO aType, void *aCtx) {
  const char *typetxt;

  switch(aType) {
    case CBSI_BAUDRATE   : typetxt = "baud"; break;
    case CBSI_DATABITS   : typetxt = "databits"; break;
    case CBSI_PARITY     : typetxt = "parity"; break;
    case CBSI_STOPBITS   : typetxt = "stopbits"; break;
    case CBSI_SERIALPORT : typetxt = "port"; break;
    case CBSI_READY      : typetxt = ""; break;
    default              : typetxt = "unknown"; break;
  }
  if (aText != NULL)
    printf("%8s : %6s\n", typetxt, aText);
} /* _portinfo() */

/** \brief Callback function for handling retrieved GPS coordinates.
  *
  * Prints the retrieved coordinates to stdout.
  * \param anUTC    the UTC time,
  * \param anNLat   the latitude, positive for North, negative for South.
  * \param anELon   the longitude, positive for East, negative for West.
  * \param anAlt    the altitude in meters.
  * \param aUTME    the UTM easting in kilometers.
  * \param aUTMN    the UTM northing in kilometers.
  * \param aCtx     casted tp struct _gpsctx. Used in order to decide, if
  *                 the coordinates changed signoficantly enough to show them.
  */
static void _gpscb(double anUTC,
                   double anNLat,
                   double anELon,
                   double anAlt,
                   double aUTME,
                   double aUTMN,
                   void   *aCtx) {
  static const double MINCOORD = 0.0001;
  static const double MINALT = 0.1;
  static const double MINUTM = 0.001;
  struct _gpsctx      *ctx;

  if (aCtx != NULL) {
    ctx = (struct _gpsctx *) aCtx;
    if (fabs(anNLat - ctx->oldlat) >= MINCOORD ||
        fabs(anELon - ctx->oldlon) >= MINCOORD ||
        fabs(anAlt - ctx->oldalt) >= MINALT ||
        fabs(aUTME - ctx->oldutme) >= MINUTM ||
        fabs(aUTMN - ctx->oldutmn) >= MINUTM) {
      printf("UTC %010.3lf lat. %lf lon. %lf alt. %lf UTM %lf %lf\n",
             anUTC, anNLat, anELon, anAlt, aUTME, aUTMN);
      fflush(stdout);
      ctx->oldlat = anNLat;
      ctx->oldlon = anELon;
      ctx->oldalt = anAlt;
      ctx->oldutme = aUTME;
      ctx->oldutmn = aUTMN;
      cbgpx_write(&(ctx->hgpx), anNLat, anELon, anAlt);
    }
  }
} /* _gpscb() */

/** \brief Prints usage info incl. information about available serial ports.
  * \param aProgName  name of this program, usually argv[0].
  */
static void _portinfo(const char *aProgName) {
  t_cbserinfo cbsi;

  cbserinitinfo(&cbsi);
  printf("Usage %s port baud databits parity stopbits\n", aProgName);
  printf("Values available for the parameters:\n");
  cbserinfo(&cbsi, _cbsicb, NULL);
  cbserinfo_end(&cbsi, 0);
} /* _portinfo() */

#define MAXBUF 127

/** \brief Makes the whole job.
  *
  * Starts reading gps and prints coordinates coming in the callback function.
  * \param aPort    the name of the serial port
  * \param aBaud    the baud rate
  * \param databits number of databits
  * \param aParity  the parity
  * \param stopbits number of stopbits
  */
static void _process(const char *aPort,
                     const char *aBaud,
                     const char *databits,
                     const char *aParity,
                     const char *stopbits) { 
  char           buf[MAXBUF+1];
  t_cbgps        hgps;
  struct _gpsctx ctx;
  char           *ptr;

  memset(&ctx, 0, sizeof(ctx));
  ctx.oldlat = 0.0;
  ctx.oldlon = 0.0;
  ctx.oldalt = 0.0;
  ctx.oldutme = 0.0;
  ctx.oldutmn = 0.0;
  cbgpsinit(&hgps);
  cbgpsopenstr(&hgps, aPort, aBaud, databits, aParity, stopbits, 
               DEFAULT_TIMEOUT,
               DEFAULT_SYSTEM,
               _gpscb,
               &ctx);
  printf("enter exit to quit or gpx file name to start writing gpx\n");
  while(fgets(buf, MAXBUF, stdin) != NULL) {
    if ((ptr = strchr(buf, '\n')) != NULL) *ptr = '\0';
    if ((ptr = strchr(buf, '\r')) != NULL) *ptr = '\0';
    if (strcmp(buf, "exit") == 0 || strcmp(buf, "EXIT") == 0) break;
    if (buf[0] != '\0') {
      cbgpx_close(&(ctx.hgpx));
      if (cbgpx_opena(&(ctx.hgpx), buf) == 0)
        printf("Can't open GPX file <%s>\n", buf);
    }
  } /* while */
  printf("Please wait ...\n"); fflush(stdout);
  cbgpx_close(&(ctx.hgpx));
  cbgpsclose(&hgps);
} /* _process() */

/** \brief The main function of the console application.
  *
  * Call cbcons without parameters for the usage info.
  * \param argc  number of command line parameters
  * \param argv  command line parameters, at least 5 required:
  *              port, baud, databits, parity, stopbits.
  */
int main(int argc, char **argv) {
  if (argc >= 5) {
    _process(argv[1], argv[2], argv[3], argv[4], argv[5]);
  }
  else {
    _portinfo(argv[0]);
  }
  return 0;
} /* main() */

