#include "cbgps.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#if defined(_WIN32)
#  include <malloc.h>
#elif defined(__linux__) || defined(sun)
#  include <alloca.h>
#endif



/** \file cbgps.c
  * \brief Implementation of GPS oriented functions.
  */

/** \brief Maximal length of a value string in NMEA GGA
  */
#define TOKENLEN 15

/** \brief PI.
  *
  * M_PI is not always defined in math.h, even for _USE_MATH_DEFINES.
  */
static const double PI = 3.14159265358979323846;

/** \brief Equatorial radius in meters, UTM.
  *
  * For systems like in t_cgps_utmsys.
  */
static const double _AA[] = {
      6378137.0,6378137.0,6378135.0,6378160.0,6378245.0,6378388.0,
      6378249.1,6378206.4,6377563.4,6377397.2,6377276.3
  };

/** \brief Polar Radius in meters, UTM.
  *
  * For systems like in t_cgps_utmsys.
  */
static const double _BB[] = {
      6356752.3142,6356752.3141,6356750.5000,6356774.7000,
      6356863.0000,6356911.9000,6356514.9000,6356583.8000,
      6356256.9000,6356079.0000,6356075.4000
  };

/** \brief Locale independend strtod.
  * \param theNum the double number resulting from the string anStr.
  * \param anStr  the string representation of the number, always a dot as
  *               decimal separator.
  * \returns != 0 if everything went OK, 0 on error.
  */
static int _listrtod(double *theNum, const char *anStr) {
  double d1, d2;
  size_t l, i;
  char   *ptr, *s, *endptr;
  int    retval;

  retval = 0;
  if (anStr != NULL && theNum != NULL && (l = strlen(anStr)) > 0) {
    s = (char *) alloca(l + 1);
    strncpy(s, anStr, l); s[l] = '\0';
    ptr = strchr(s, '.');
    if (ptr != NULL) {
      *ptr = '\0';
      ptr++;
      d1 = (double) strtol(s, &endptr, 10);
      if (endptr == NULL || *endptr == '\0') {
        l = strlen(ptr);
        if (l > 0) {
          d2 = (double) strtol(ptr, &endptr, 10);
          if (endptr == NULL || *endptr == '\0') {
            for(i = 0; i < l; i++) d2 *= 0.1;
          }
        }
        else d2 = 0.0;
        *theNum = d1 + d2;
      }
    }
    else {
      *theNum = (double) strtol(s, &endptr, 10);
    }
    if (endptr == NULL || *endptr == '\0') retval = 1;
  }
  return retval;
} /* _listrtod */

/** \brief Converts degrees in NMEA GGA format to decimal degrees.
  * \param dms  the degrees in format used in NMEA sequences, eg.
  *             1234.8765 means 12 degress, 34.8765 minutes,
  *             equaling 12 def 34 min and 52.59 sec or
  *             12.581275 deg.
  * \returns    decimal degrees, 12.581275 for the example above.
  */
static double _dms2dec(double dms) {
  double deg, min, deg100, part;

  deg = floor(0.01 * dms);
  deg100 = 100 * deg;
  min = floor(dms - deg100);
  part = dms - deg100 - min;
  return deg + (min + part)/ 60;
} /* _dms2dec() */

/** \brief Converts latitude and longitude to UTM northing and easting.
  *
  * Based on the theoretical description and an Excel sheet from
  * Prof. Steve Dutch, see:
  * http://www.uwgb.edu/DutchS/UsefulData/UTMFormulas.HTM and
  * http://www.uwgb.edu/DutchS/UsefulData/UTMConversions1.xls
  * \param utmnorthing out, the resulting UTM northing in km.
  * \param utmeasting  out, the resulting UTM easting in km.
  * \param n_lat       North latitude (negative for South) in decimal degress.
  * \param e_long      East longitude (negative for West) in decimal degrees.
  * \param system      the system used for recomputing, WGS84, GRS80 etc.
  */
static void _toUTM(double *utmnorthing,
                   double *utmeasting,
                   double n_lat,
                   double e_long,
                   t_cbgps_utmsys system) {
  const double k0 = 0.9996;
  const double sin1sec1 = 4.8481368110763677e-006; /* sin(1") */
  const double sin1sec2 = 2.3504430538913731e-011; /* sin(1")^2 */
  const double sin1sec3 = 1.1395269491909521e-016; /* sin(1")^3 */
  const double sin1sec4 = 5.5245825495862045e-022; /* sin(1")^4 */
  double       a, b, a2, b2, n, n_1, n2, n3, n4, n5, n2_3, n4_5,
               aprim, bprim, cprim, dprim, eprim, s,
               lat, lat2, lat4, lat6, lat8, k1, k2, k3, k4, k5,
               sinlat, coslat, coslat2, coslat3, coslat4, tanlat, tanlat2,
               nu, e, e2, e4, p, p2, p3, p4;
  int          idx, lonzone, lonzonecm;

  idx = (int) system;
  a = _AA[idx];
  b = _BB[idx];
  a2 = a * a; b2 = b * b;
  e = sqrt((a2 - b2) / a2);
  e2 = e * e; e4 = e2 * e2;
  n = (a - b) / (a + b);
  n2 = n * n; n3 = n2 * n; n4 = n3 * n; n5 = n4 * n;
  n2_3 = n2 - n3; n4_5 = n4 - n5;
  n_1 = 1.0 - n;
  aprim = a * (n_1 + 1.25 * n2_3 + 1.265625 * n4_5);
  bprim = 1.5 * a * n * (n_1 + 0.875 * n2_3 + 0.859375 * n4_5);
  cprim = 0.9375 * a * n2 * (n_1 + 0.75 * n2_3);
  dprim = 35.0 * a * n3 * (n_1 + 11.0 * n2_3 / 16.0) / 48.0;
  eprim = 315.0 * a * n4 * n_1 / 51.0;
  lat = n_lat * PI / 180.0;
  lat2 = lat + lat;
  lat4 = lat2 + lat2;
  lat6 = lat4 + lat2;
  lat8 = lat4 + lat4;
  sinlat = sin(lat);
  coslat = cos(lat);
  coslat2 = coslat * coslat;
  coslat3 = coslat2 * coslat;
  coslat4 = coslat2 * coslat2;
  tanlat = tan(lat);
  tanlat2 = tanlat * tanlat;
  lonzone = 31 + (int) floor(e_long / 6);
  lonzonecm = 6 * lonzone - 183;
  p = 0.36 * (e_long - lonzonecm); /* (lon - lonzonecm) * 3600 / 10000 */
  p2 = p * p; p3 = p2 * p; p4 = p2 * p2;
  s = aprim * lat - bprim * sin(lat2) + cprim * sin(lat4) -
      dprim * sin(lat6) + eprim * sin(lat8);
  nu = a / sqrt(1.0 - e2 * sinlat * sinlat);
  k1 = s * k0;
  k2 = nu * sinlat * coslat * sin1sec2 * k0 * 50000000;
  k3 = k0 * sin1sec4 * nu * sinlat * coslat3 *
       (5.0 - tanlat * tanlat + 9.0 * e2 * coslat2 + 4.0 * e4 * coslat4) *
       1e16/ 24.0;
  k4 = nu * coslat * sin1sec1 * k0 * 10000;
  k5 = 1000000000000.0 * k0 * sin1sec3 * nu * coslat3 *
       (1.0 - tanlat2 + e2 * coslat2) / 6.0;
  if (utmnorthing != NULL) *utmnorthing = 0.001 * (k1 + k2 * p2 + k3 * p4);
  if (utmeasting != NULL) *utmeasting = 0.001 * (500000.0 + k4 * p + k5 * p3);
} /* _toUTM() */

/** \brief The implementation of the callback function for reading lines
  *        of text from the serial port.
  *
  * Tries to parse the given lines of text using cbgps_gga(). It it succeeds,
  * the it calls the callback function passing the extracted UTC time and 
  * coordinates as parameters to it. The pointer to the callback function is
  * expected in the context variable.
  * \param aLine   the line of text.
  * \param aLen    the length of the line of text.
  * \param aCtx    context casted to t_cbgps.
  */
static void _rdfkt(const char *aLine, size_t aLen, void *aCtx) {
  t_cbgps  *ctx;
  double   utc, nlat, elon, alt, utmn, utme;

  if (aCtx != NULL) {
    ctx = (t_cbgps *) aCtx;
    if (cbgps_gga(aLine,
                  ctx->system,
                  &utc,
                  &nlat,
                  &elon,
                  &alt,
                  &utmn,
                  &utme) != 0) {
      if (ctx->rdCb != NULL)
        ctx->rdCb(utc, nlat, elon, alt, utmn, utme, ctx->ctx);
    }
  }
} /* _rdfkt() */

int cbgps_gga(const char     *aGGASeq,
              t_cbgps_utmsys aUTMSystem,
              double         *theUTC,
              double         *theNLat,
              double         *theELong,
              double         *theAlt,
              double         *theUTMN,
              double         *theUTME) {
  int  retval;

  retval = 0;
  if (aGGASeq != NULL && theUTC != NULL &&
      theNLat != NULL && theELong != NULL && theAlt != NULL &&
      theUTMN != NULL && theUTME != NULL) {
    *theUTC = 0.0;
    *theNLat = 0.0;
    *theELong = 0.0;
    *theAlt = 0.0;
    *theUTMN = 0.0;
    *theUTME = 0.0;
    if (aGGASeq[0] == '$' &&
        aGGASeq[1] != '\0' &&
        aGGASeq[2] != '\0' &&
        aGGASeq[3] == 'G' &&
        aGGASeq[4] == 'G' &&
        aGGASeq[5] == 'A' &&
        aGGASeq[6] == ',') {
      char       token[TOKENLEN+1];
      double     d;
      char       *ptok;
      const char *pgga;
      int        itok, ilen, isok;
 
      itok = 0;
      ptok = token;
      ilen = 0;
      for(pgga = aGGASeq + 7; *pgga != '\0'; pgga++) {
        isok = 0;
        if (ilen >= TOKENLEN) break;
        if (*pgga == ',' ) {
          *ptok = '\0';
          switch(itok) {
            case 0 : /* UTC time */
              isok = _listrtod(&d, token);
              if (isok != 0) *theUTC = d;
              break;
            case 1 : /* latitude */
              isok = _listrtod(&d, token);
              if (isok != 0) *theNLat = _dms2dec(d);
              break;
            case 2 : /* N or S (for North or South) */
              if (token[0] == 'S' && token[1] == '\0') {
                *theNLat = - (*theNLat);
                isok = 1;
              }
              else if (token[0] == 'N' && token[1] == '\0') isok = 1;
              break;
            case 3 : /* longitude */
              isok = _listrtod(&d, token);
              if (isok != 0) *theELong = _dms2dec(d);
              break;
            case 4 : /* E or W (for East or West) */
              if (token[0] == 'W' && token[1] == '\0') {
                *theELong = - (*theELong);
                isok = 1;
              }
              else if (token[0] == 'E' && token[1] == '\0') isok = 1;
              break;
            case 8 : /* altitude */
              isok = _listrtod(theAlt, token);
              break;
            default: /* ignored */
              isok = 1;
              break;
          } /* switch(itok) */
          if (isok == 0 || itok >= 8) break; /* stop parsing not interesting stuff */
          ptok = token;
          ilen = 0;
          itok++;
        }
        else {
          *ptok = *pgga;
          ptok++;
          ilen++;
        }
      } /* for */
      if (isok != 0) {
        _toUTM(theUTMN, theUTME, *theNLat, *theELong, aUTMSystem);
        retval = 1;
      }
    } /* if */
  } /* if */
  return retval;
} /* cbgps_gga() */

int cbgpsopen(t_cbgps            *theHandle,
              const char         *aComPort,
              const t_cbserparam *aSpeed,
              const t_cbserparam *databits,
              const t_cbserparam *aParity,
              const t_cbserparam *stopbits,
              int                aTimeout,
              t_cbgps_utmsys     aSystem,
              t_cbgpscb          rdCb,
              void               *aCtx) {
  int  retval;

  retval = 0;
  cbgpsinit(theHandle);
  if (theHandle != NULL) {
    theHandle->rdCb = rdCb;
    theHandle->system = aSystem;
    theHandle->ctx = aCtx;
    retval = cbseropen(&(theHandle->hcom),
                       aComPort,
                       aSpeed,
                       databits,
                       aParity,
                       stopbits,
                       aTimeout,
                       _rdfkt,
                       theHandle);
  }
  return retval;
} /* cbgpsopen() */

int cbgpsopenstr(t_cbgps        *theHandle,
                 const char     *aComPort,
                 const char     *aSpeed,
                 const char     *databits,
                 const char     *aParity,
                 const char     *stopbits,
                 int            aTimeout,
                 t_cbgps_utmsys aSystem,
                 t_cbgpscb      rdCb,
                 void           *aCtx) {
  int retval;

  retval = 0;
  cbgpsinit(theHandle);
  if (theHandle != NULL) {
    theHandle->rdCb = rdCb;
    theHandle->system = aSystem;
    theHandle->ctx = aCtx;
    retval = cbseropenstr(&(theHandle->hcom),
                          aComPort,
                          aSpeed,
                          databits,
                          aParity,
                          stopbits,
                          aTimeout,
                          _rdfkt,
                          theHandle);
  }
  return retval;
} /* cbgpsopenstr() */

int cbgpsopenwstr(t_cbgps        *theHandle,
                  const wchar_t  *aComPort,
                  const wchar_t  *aSpeed,
                  const wchar_t  *databits,
                  const wchar_t  *aParity,
                  const wchar_t  *stopbits,
                  int            aTimeout,
                  t_cbgps_utmsys aSystem,
                  t_cbgpscb      rdCb,
                  void           *aCtx) {
  int retval;

  retval = 0;
  cbgpsinit(theHandle);
  if (theHandle != NULL) {
    theHandle->rdCb = rdCb;
    theHandle->system = aSystem;
    theHandle->ctx = aCtx;
    retval = cbseropenwstr(&(theHandle->hcom),
                           aComPort,
                           aSpeed,
                           databits,
                           aParity,
                           stopbits,
                           aTimeout,
                           _rdfkt,
                           theHandle);
  }
  return retval;
} /* cbgpsopenwstr() */

int cbgpsclose(t_cbgps *aHandle) {
  int retval;

  retval = 0;
  if (aHandle != NULL) {
    retval = cbserclose(&(aHandle->hcom));
  }
  return retval;
} /* cbgpsclose() */

void cbgpsinit(t_cbgps *theHandle) {
  if (theHandle != NULL) {
    cbserinitcom(&(theHandle->hcom));
    theHandle->rdCb = NULL;
    theHandle->system = WGS84;
  }
} /* cbgpsinit() */

