#include "cbgpx.h"
#ifdef FOR_MS_WINDOWS
# include <windows.h>
#else
# include <time.h>
#endif
#include <string.h>
#include <stdlib.h>

static const double MAXDOUBLE = 1e10;

#if defined(FOR_MS_WINDOWS) && !defined(__CEGCC__)
# define PRTF _snprintf
#else
# define PRTF snprintf
#endif

#define MAXSTRLEN 31

#ifndef FILENAME_MAX
/* not defined on windows ce */
# ifdef MAX_PATH
#   define FILENAME_MAX MAX_PATH
# else
#   define FILENAME_MAX 254
# endif
#endif

/** \struct _gpxtime
  * \brief Timestamp description.
  */
struct _gpxtime {
  /** \brief Year, eg. 2009 */
  short year;
  /** \brief Month, 1-12 */
  short month;
  /** \brief Day, 1-31 */
  short day;
  /** \brief Hour, 0-23 */
  short hour;
  /** \brief Minute, 0-59 */
  short min;
  /** \brief Second, 0-59 */
  short sec;
};

/** \brief Sets default values to all fields of the given handle.
  * \param aHandle  pointer to the handle to reset
  */
static void _reset_handle(t_cbgpx *aHandle) {
  if (aHandle != NULL) {
    aHandle->f = NULL;
    aHandle->minlat = MAXDOUBLE;
    aHandle->maxlat = -MAXDOUBLE;
    aHandle->minlon = MAXDOUBLE;
    aHandle->maxlon = -MAXDOUBLE;
    aHandle->idx = 0;
    cbthr_init_cs(&(aHandle->cs));
  }
} /* _reset_handle() */

/** \brief Converts the given double number to a string.
  *
  * Ensures, that dot is used as decimal separator rather than the
  * character set in the current locales.
  * \param theBuf  the buffer, where to store the resulting string
  * \param aLen    length of the buffer, the buffer pointed by theBuf
  *                must be able to hold at least aLen+1 characters.
  * \param aNum    the number to convert.
  * \return        theBuf
  */
const char *_dbl2str(char *theBuf, size_t aLen, double aNum) {
  char *c;

  if (theBuf != NULL) {
    PRTF(theBuf, aLen, "%lf", aNum);
    c = strchr(theBuf, ',');
    if (c != NULL) *c = '.';
  }
  return theBuf;
} /* _dbl2str() */

/** \brief Estimates the current system time.
  * \param theBuf the resulting time structure with the current time.
  * \return       theTM
  */
const struct _gpxtime *_currtm(struct _gpxtime *theTM) {
#ifdef FOR_MS_WINDOWS
  SYSTEMTIME st;
#else
  struct tm  ttm;
  time_t     tt;
#endif
  if (theTM != NULL) {
    memset(theTM, 0, sizeof(struct _gpxtime));
#ifdef FOR_MS_WINDOWS
    GetSystemTime(&st);
    theTM->year = (short) (st.wYear);
    theTM->month = (short) (st.wMonth);
    theTM->day = (short) (st.wDay);
    theTM->hour = (short) (st.wHour);
    theTM->min = (short) (st.wMinute);
    theTM->sec = (short) (st.wSecond);
#else
    time(&tt);
    gmtime_r(&tt, &ttm);
    theTM->year = (short) (ttm.tm_year + 1900);
    theTM->month = (short) (ttm.tm_mon + 1);
    theTM->day = (short) (ttm.tm_mday);
    theTM->hour = (short) (ttm.tm_hour);
    theTM->min = (short) (ttm.tm_min);
    theTM->sec = (short) (ttm.tm_sec);
#endif
  }
  return theTM;
} /* _currtm() */

short cbgpx_isvalid(const t_cbgpx* aHandle) {
  if (aHandle != NULL && aHandle->f != NULL) return 1;
  return 0;
} /* cbgpx_isvalid() */

short cbgpx_opena(t_cbgpx* theHandle, const char *aFileName) {
  struct _gpxtime ttm;
  FILE            *f;
  short           retval;

  retval = 0;
  if (theHandle != NULL) {
    _reset_handle(theHandle);
    if (aFileName != NULL) {
      f = fopen(aFileName, "wt");
      if (f != NULL) {
        _currtm(&ttm);
        if (fprintf(f, "<?xml version=\"1.0\"?>\n") > 0 &&
            fprintf(f,
                    "<gpx xmlns=\"http://www.topografix.com/GPX/1/0\">\n") > 0 &&
            fprintf(f,
                    "  <time>%04d-%02d-%02dT%02d:%02d:%02dZ</time>\n",
                    ttm.year,
                    ttm.month,
                    ttm.day,
                    ttm.hour,
                    ttm.min,
                    ttm.sec) > 0) {
          if (cbthr_enter_cs(&(theHandle->cs)) == 0) {
            theHandle->f = f;
            if (cbthr_leave_cs(&(theHandle->cs)) == 0) retval = 1;
          }
        } /* if (fprintf ...) */
      } /* if (f != NULL) */
    } /* if (aFileName != NULL) */
  } /* if (theHandle != NULL) */
  return retval;
} /* cbgpx_opena() */

short cbgpx_openw(t_cbgpx* theHandle, const wchar_t *aFileName) {
  char  filename[FILENAME_MAX+1];
  short retval;

  retval = 0;
  if (aFileName != NULL) {
    if (wcstombs(filename, aFileName, FILENAME_MAX) != (size_t) -1) {
      filename[FILENAME_MAX] = '\0';
      retval = cbgpx_opena(theHandle, filename);
    }
  }
  return retval;
} /* cbgpx_openw() */

void cbgpx_close(t_cbgpx *aHandle) {
  char minlat[MAXSTRLEN+1], minlon[MAXSTRLEN+1],
       maxlat[MAXSTRLEN+1], maxlon[MAXSTRLEN+1];

  if (cbgpx_isvalid(aHandle) != 0) {
    cbthr_enter_cs(&(aHandle->cs));
    _dbl2str(minlat, MAXSTRLEN, aHandle->minlat);
    _dbl2str(minlon, MAXSTRLEN, aHandle->minlon);
    _dbl2str(maxlat, MAXSTRLEN, aHandle->maxlat);
    _dbl2str(maxlon, MAXSTRLEN, aHandle->maxlon);
    fprintf(aHandle->f,
            "  <bounds minlat=\"%s\" minlon=\"%s\" maxlat=\"%s\" maxlon=\"%s\"/>\n",
            minlat, minlon, maxlat, maxlon);
    fprintf(aHandle->f, "</gpx>\n");
    fclose(aHandle->f);
    cbthr_leave_cs(&(aHandle->cs));
    cbthr_delete_cs(&(aHandle->cs));
  }
  _reset_handle(aHandle);
} /* cbgpx_close() */

short cbgpx_write(t_cbgpx *aHandle, double aLat, double aLon, double anAlt) {
  char            lat[MAXSTRLEN+1], lon[MAXSTRLEN+1], alt[MAXSTRLEN+1];
  struct _gpxtime ttm;
  short           retval;

  retval = 0;
  if (cbgpx_isvalid(aHandle) != 0) {
    if (cbthr_enter_cs(&(aHandle->cs)) == 0) {
      if (cbgpx_isvalid(aHandle) != 0) {
          _dbl2str(lat, MAXSTRLEN, aLat);
        _dbl2str(lon, MAXSTRLEN, aLon);
        _dbl2str(alt, MAXSTRLEN, anAlt);
        _currtm(&ttm);
        if (fprintf(aHandle->f, "  <wpt lat=\"%s\" lon=\"%s\">\n", lat, lon) > 0 &&
            fprintf(aHandle->f, "    <time>%04d-%02d-%02dT%02d:%02d:%02dZ</time>\n",
                    ttm.year,
                    ttm.month,
                    ttm.day,
                    ttm.hour,
                    ttm.min,
                    ttm.sec) > 0 &&
            fprintf(aHandle->f, "    <ele>%s</ele>\n", alt) > 0 &&
            fprintf(aHandle->f, "    <name>%d</name>\n", aHandle->idx) > 0 &&
            fprintf(aHandle->f, "    <desc>%d</desc>\n", aHandle->idx) > 0 &&
            fprintf(aHandle->f, "    <sym>Dot</sym>\n") > 0 &&
            fprintf(aHandle->f, "    <type>Dot</type>\n") > 0 &&
            fprintf(aHandle->f, "  </wpt>\n") > 0) {
          aHandle->idx++;
          if (aLat > aHandle->maxlat) aHandle->maxlat = aLat;
          if (aLat < aHandle->minlat) aHandle->minlat = aLat;
          if (aLon > aHandle->maxlon) aHandle->maxlon = aLon;
          if (aLon < aHandle->minlon) aHandle->minlon = aLon;
          retval = 1;
        }
      } /* if (cbgpx_isvalid(aHandle) != 0) */
      if (cbthr_leave_cs(&(aHandle->cs)) != 0) retval = 0;
    } /* if (cbthr_enter_cs(&(aHandle->cs)) == 0) */ 
  } /* if (aHandle != NULL) */
  return retval;
} /* cbgpx_write() */

