#ifndef __CBGPX_H_
#define __CBGPX_H_

#include "cbthr.h"
#include <stdio.h>
#include <wchar.h>

/** \file cbgpx.h
  * \brief Set of functions for creating GPX files.
  *
  * See also GPX: The GPS exchange format , http://www.topografix.com/gpx.asp.
  *
  * Example: the following code:
  * \code
  * #include "cbgpx.h"
  *
  * int main() {
  *   t_cbgpx  hgpx;
  *
  *   if (cbgpx_opena(&hgpx, "test.gpx") != 0) {
  *     cbgpx_write(&hgpx, 49.2, 11.7, 403.6);
  *     ...
  *     cbgpx_write(&hgpx, 49.5, 12.2, 398.3);
  *     cbgpx_close(&hgpx);
  *   }
  * }
  * \endcode
  * produces the following GPX file in "test.gpx":
  * \code
  * <?xml version="1.0"?>
  * <gpx xmlns="http://www.topografix.com/GPX/1/0">
  *   <time>2008-11-23T16:40:07Z</time>
  *   <wpt lat="49.2" lon="11.7">
  *     <ele>403.6</ele>
  *     <time>2008-11-23T16:46:09Z</time>
  *     <name>0</name>
  *     <desc>0</desc>
  *     <sym>Dot</sym>
  *     <type>Dot</type>
  *   </wpt>
  *   <!-- ... -->
  *   <wpt lat="49.5" lon="12.2">
  *     <ele>398.3</ele>
  *     <time>2008-11-23T17:05:51Z</time>
  *     <name>275</name>
  *     <desc>275</desc>
  *     <sym>Dot</sym>
  *     <type>Dot</type>
  *   </wpt>
  *   <bounds minlat="49.2" minlon="11.7" maxlat="49.5" maxlon="12.2" />
  * </gpx>
  * \endcode
  */

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  /** \brief Critical secrion/mutex for synchronizing file operations.
    */
  t_cbthrcs cs;
  /** \brief Minimum latitude until now.
    *
    * Negative and positive decimal degrees.
    */
  double    minlat;
  /** \brief Maximum latitude until now.
    *
    * Negative and positive decimal degrees.
    */
  double    maxlat;
  /** \brief Minimum longitude until now.
    *
    * Negative and positive decimal degrees.
    */
  double    minlon;
  /** \brief Maximum longitude until now.
    *
    * Negative and positive decimal degrees.
    */
  double    maxlon;
  /** \brief Handle of the created GPX file.
    */
  FILE      *f;
  /** \brief Sequential number of the next entry.
    *
    * Starting with 0.
    */
  int       idx;
} t_cbgpx;

/** \brief Checks, if the given handle of the GPX file is valied.
  * \param aHandle  pointer to the handle to check
  * \return         != 0 if the handle is valid, 0 else.
  */
short cbgpx_isvalid(const t_cbgpx* aHandle);

/** \brief Opens the given file as a GPX file, ANSI file name.
  *
  * The function opens the given file and writes the following initial
  * information into it:
  * \code
  * <?xml version="1.0"?>
  * <gpx xmlns="http://www.topografix.com/GPX/1/0">
  *   <time>[current time]</time>
  * \endcode
  * where [current time] is the current system time in format
  * yyyy-MM-ddTHH:mm:ssZ
  * \param theHandle  the resulting handle of the GPX file.
  * \param aFileName  the name of the file to open.
  * \return != 0 if everything went OK, 0 on error.
  */
short cbgpx_opena(t_cbgpx* theHandle, const char *aFileName);

/** \brief Opens the given file as a GPX file, wide char file name.
  *
  * Converts the given file name into multibyte and uses
  * cbgpx_opena() in order to open the GPX file.
  * \param theHandle  the resulting handle of the GPX file.
  * \param aFileName  the name of the file to open.
  * \return != 0 if everything went OK, 0 on error.
  */
short cbgpx_openw(t_cbgpx* theHandle, const wchar_t *aFileName);

/** \brief Closes the GPX file.
  *
  * The function writes the following closing information into the GPX file
  * \code
  * \endcode
  * and closes the file handle. Afterwards it resets the handle of the GPX file.
  * \param aHandle  the handle of the GPX file to close.
  */
void cbgpx_close(t_cbgpx *aHandle);

/** \brief Writes the given coordinates into the GPX file.
  *
  * The function writes the following information into the GPX file:
  * \code
  *   <wpt lat="[latitude]" lon="[longitude]">
  *     <ele>[altitude]</ele>
  *     <time>[current time]</time>
  *     <name>0</name>
  *     <desc>0</desc>
  *     <sym>Dot</sym>
  *     <type>Dot</type>
  *   </wpt>
  * \endcode
  * where:
  * -# [latitude] is the given latitude aLat with a dot as decimal point
  * -# [longitude] is the given longitude aLon with a dot as decimal point
  * -# [altitude] is the given altitude anAlt with a dot as decimal point
  * -# long is the given latitude with dot as the decimal point separator
  * -# [current time] is the current system time in format
  *     yyyy-MM-ddTHH:mm:ssZ
  *
  * \param aHandle the GPX file handle.
  * \param aLat    the latitude, positive for North, negative for South,
  *                decimal degrees.
  * \param aLon    the longitude, positive for East, negative for West,
  *                decimal degrees.
  * \param anAlt   the altitude in meters.
  * \return != 0 if everything went OK, 0 on error.
  */
short cbgpx_write(t_cbgpx *aHandle, double aLat, double aLon, double anAlt);

#ifdef __cplusplus
}
#endif

#endif

