#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <shlobj.h>
#include <math.h>
#include <stdlib.h>
#include "cbgps.h"
#include "cbrc.h"
#include "cbgpx.h"
#ifdef UNICODE
# include <wchar.h>
#else
# include <string.h>
#endif

#ifdef __CEGCC__
extern void SystemIdleTimerReset(void);
extern int _snwprintf(wchar_t*, size_t, const wchar_t*, ...);
#endif

/** \file cbgui.c
  * \brief WIN32 GUI
  */

/** \brief Maximal string length, withou tthe trailing 0.
  */
#define MAX_STR_LEN 127

/** \brief Length of an configuration item.
  *
  * A configuration item is for example the name of the COM port, baud rate,
  * number of databits etc.
  */
#define CFG_STR_LEN  31

/** \brief Length of the string for a value retrieved by the GPS callback.
  */
#define VAL_STR_LEN 15
  
#ifdef UNICODE
# ifdef __WINE__
#  define SPRTF swprintf
# else
#  define SPRTF _snwprintf
# endif
#else
# ifdef __WINE__
#  define SPRTF snprintf
# else
#  define SPRTF _snprintf
# endif
#endif

#define MENU_START_ENABLED    1
#define MENU_STOP_ENABLED     2
#define MENU_CONFIG_ENABLED   4
#define MENU_GPXSTART_ENABLED 8
#define MENU_GPXSTOP_ENABLED 16
#define MENU_EXIT_ENABLED    32
#define MENU_ALL_ENABLED     ~0

/** \brief Message sent to the configuration window after the serial
  *        info callback is ready.
  */
#define WM_CBSIREADY WM_USER + 1234

/** \brief Message sent to the main window, if new coordinates arrived.
  *
  * Message is sent in _gpscb() to the main window with a bitwise
  * combination of the COORD_ constants in wParam.
  */
#define WM_NEWCOORD WM_USER + 1235

/** \brief Bitwise information, that UTC time changed
  */
#define COORD_UTC 0x0001

/** \brief Bitwise information, that the altitude changed
  */
#define COORD_ALT 0x0002

/** \brief Bitwise information, that latitude changed
  */
#define COORD_LAT 0x0004

/** \brief Bitwise information, that longitude changed
  */
#define COORD_LON 0x0008

/** \brief Bitwise information, that UTM northing changed
  */
#define COORD_UTMN 0x0010

/** \brief Bitwise information, that UTM easting changed
  */
#define COORD_UTME 0x0020

/** \struct _config
  * \brief Present configuration of the program.
  */
struct _config {
  /** \brief Name of the serial port. */
  TCHAR portname[CFG_STR_LEN+1];
  /** \brief Name for the baudrate. */
  TCHAR baudrate[CFG_STR_LEN+1];
  /** \brief Name for the number of databits. */
  TCHAR databits[CFG_STR_LEN+1];
  /** \brief Name for the parity. */
  TCHAR parity[CFG_STR_LEN+1];
  /** \brief Name for number of stopbits. */
  TCHAR stopbits[CFG_STR_LEN+1];
  /** \brief ID of the selection in the combo box for portnames. */
  short portid;
  /** \brief ID of the selection in the combo box for baudrate names. */
  short baudid;
  /** \brief ID of the selection in the combo box for databits names. */
  short databitsid;
  /** \brief ID of the selection in the combo box for parity names. */
  short parityid;
  /** \brief ID of the selection in the combo box for stopbit names. */
  short stopbitsid;
};

/** \struct _cbsientry
  * \brief One configuration entry in the list of entries set in the
  *        serial port information callback.
  */
struct _cbsientry {
  /** \brief The value of the entry.
    *
    * Eg. serial port name, name of the bauf, parity, number of databits
    * or stopbits.
    */
  TCHAR val[CFG_STR_LEN+1];
  /** \brief Type of the value in val.
    */
  CBSERINFO typ;
  /** \brief Pointer to the next element in the list.
    */
  struct _cbsientry *next;
};

/** \struct _cbsictx
  * \brief Context for the serial port information callback.
  */
struct _cbsictx {
  /** \brief Begin of the list of serial port information entries.
    */
  struct _cbsientry *first;
  /** \brief Handle of the configuration dialog window.
    */
  HWND              hwnd;
};

/** \struct _gpscbcontrols
  * \brief Window handles of controls to fill with current coordinates.
  */
struct _gpscbcontrols {
  /** \brief String buffer for a value of the UTC time */
  TCHAR          sutc[VAL_STR_LEN+1];
  /** \brief String buffer for a value of the latitude */
  TCHAR          slat[VAL_STR_LEN+1];
  /** \brief String buffer for a value of the longitude */
  TCHAR          slon[VAL_STR_LEN+1];
  /** \brief String buffer for a value of the altitude */
  TCHAR          salt[VAL_STR_LEN+1];
  /** \brief String buffer for a value of the utm easting */
  TCHAR          sutme[VAL_STR_LEN+1];
  /** \brief String buffer for a value of the utm northing */
  TCHAR          sutmn[VAL_STR_LEN+1];
  /** \brief Handle of the GPX file.
    */
  t_cbgpx        hgpx;
  /** \brief Critical section for synchronizing operations on
    *        contents of this structure.
    */
  t_cbthrcs      cs;
  /** \brief Previous UTC time.
    *
    * Current UTC time is inserted into the control if it
    * changes by at least 1 second.
    */
  double         oldutc;
  /** \brief Previous latitude.
    *
    * Current latitude is inserted into the control if it
    * changes by at least 0.0001 degree.
    */
  double         oldlat;
  /** \brief Previous longitude.
    *
    * Current longitude is inserted into the control if it
    * changes by at least 0.0001 degree.
    */
  double         oldlon;
  /** \brief Previous altitude.
    *
    * Current altitude is inserted into the control if it
    * changes by at least 0.1 m.
    */
  double         oldalt;
  /** \brief Previous UTM easting.
    *
    * Current UTM easting is inserted into the control if it
    * changes by at least 1 m.
    */
  double         oldutme;
  /** \brief Previous UTM northing.
    *
    * Current UTM northing is inserted into the control if it
    * changes by at least 1 m.
    */
  double         oldutmn;
  /** \brief Handle of the main windows.
    */
  HWND           hwnd;
};

static const TCHAR _APPNAME[] = TEXT("cb");
static TCHAR       _szQuit[MAX_STR_LEN+1],
                   _szOpenErr[MAX_STR_LEN+1],
                   _szCloseErr[MAX_STR_LEN+1],
                   _szFmtUTC[VAL_STR_LEN+1],
                   _szFmtNLat[VAL_STR_LEN+1],
                   _szFmtSLat[VAL_STR_LEN+1],
                   _szFmtELon[VAL_STR_LEN+1],
                   _szFmtWLon[VAL_STR_LEN+1],
                   _szFmtAlt[VAL_STR_LEN+1],
                   _szFmtUTME[VAL_STR_LEN+1],
                   _szFmtUTMN[VAL_STR_LEN+1],
                   _szGPXFilter[MAX_STR_LEN+1];

/** \brief Module for loading localized resources.
  *
  * Estimated using _nlsresmod().
  */
static HMODULE _hresmod;

/** \brief Pronts last system error in a message box.
  */
static void _fmtErrMsg() {
  TCHAR errstr[MAX_STR_LEN+1];

  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
                NULL,
                GetLastError(),
                0,
                errstr,
                MAX_STR_LEN,
                NULL);
  MessageBox(NULL, errstr, _APPNAME, MB_ICONERROR);
} /* _fmtErrMsg() */

/** \brief Releases memory occupied by the list of serial port informations.
  *
  * Such a list is build in _cbserinfocb() and it should be released, if
  * not needed anymore, eg. after the message WM_CBSIREADY has been sent.
  * \param first  begin of the list to release
  */ 
static void _cbsientry_release(struct _cbsientry *first) {
  struct _cbsientry *c, *ch;

  if (first != NULL) {
    for (c = first; c != NULL;) {
      ch = c->next;
      free(c);
      c = ch;
    }
    ZeroMemory(first, sizeof(struct _cbsientry));
  }
} /* _cbsientry_release() */

/** \brief Callback function for serial port information.
  *
  * Called for each available serial port parameter item: serial port name,
  * baud, number of databits, parity, stopbits.
  * \param aName  the name of the setting, eg. name of the serial port, number
  *               of databits, stopbits as string, parity or baudrate name.
  *               It may be also NULL!
  * \param aType  type of the information in aName. For CBSI_BAUDRATE,
  *               CBSI_DATABITS, CBSI_PARITY, CBSI_STOPBITS, CBSI_SERIALPORT
  *               the values of aName and aType is added to the list of
  *               values.
  *               For CBSI_READY the configuration window is informed
  *               (PostMessage, WM_CBSIREADY), that it may put the
  *               informations into the controls.
  * \param aCtx   casted to struct _cbsictx.
  */
static void _cbserinfocb(const char *aName, CBSERINFO aType, void *aCtx) {
  struct _cbsictx   *c;
  struct _cbsientry *cnew;

  if (aCtx != NULL) {
    switch(aType) {
      case CBSI_SERIALPORT :
      case CBSI_BAUDRATE :
      case CBSI_DATABITS :
      case CBSI_PARITY :
      case CBSI_STOPBITS :
        if (aCtx != NULL && aName != NULL) {
          cnew = (struct _cbsientry *) calloc(1, sizeof(struct _cbsientry));
          if (cnew != NULL) {
#ifdef UNICODE
            mbstowcs(cnew->val, aName, CFG_STR_LEN);
#else
            strncpy(cnew->val, aName, CFG_STR_LEN);
#endif
            cnew->val[CFG_STR_LEN] = TEXT('\0');
            cnew->typ = aType;
            c = (struct _cbsictx *) aCtx;
            cnew->next = c->first;
            c->first = cnew;
          } /* if (cnew != NULL) */
        } /* if (aCtx != NULL && aName != NULL) */
        break;
      case CBSI_READY :
        if (aCtx != NULL) {
          c = (struct _cbsictx *) aCtx;
          PostMessage(c->hwnd, WM_CBSIREADY, 0, 0);
        }
        break;
      default :
        break;
    }
  } /* if (aCtx != NULL) */
} /* _cbserinfocb() */

/** \brief Callback routine retrieving current coordinates.
  *
  * Prepares WM_NEWCOORD message and new coordinates in text format in
  * the context. Sending WM_NEWCOORD to the main window causes showing
  * new coordinates in the controls.
  * \param anUTC    the UTC time,
  * \param anNLat   the latitude, positive for North, negative for South.
  * \param anELon   the longitude, positive for East, negative for West.
  * \param anAlt    the altitude in meters.
  * \param aUTMN    the UTM northing in kilometers.
  * \param aUTME    the UTM easting in kilometers.
  * \param aCtx     casted to struct _gpscbcontrols.
  */
static void _gpscb(double anUTC,
                   double anNLat,
                   double anELon,
                   double anAlt,
                   double aUTMN,
                   double aUTME,
                   void   *aCtx) {
  static const double MINUTC = 1.0;
  static const double MINCOORD = 0.0001;
  static const double MINALT = 0.1;
  static const double MINUTM = 0.001;
  struct _gpscbcontrols *gpsc;
  DWORD                 wparam;
  short                 dogpx;

  if (aCtx != NULL) {
    dogpx = 0;
    wparam = 0;
    gpsc = (struct _gpscbcontrols *) aCtx;
    if (fabs(anUTC - gpsc->oldutc) >= MINUTC)    wparam |= COORD_UTC;
    if (fabs(anNLat - gpsc->oldlat) >= MINCOORD) wparam |= COORD_LAT;
    if (fabs(anELon - gpsc->oldlon) >= MINCOORD) wparam |= COORD_LON;
    if (fabs(anAlt - gpsc->oldalt) >= MINALT)    wparam |= COORD_ALT;
    if (fabs(aUTME - gpsc->oldutme) >= MINUTM)   wparam |= COORD_UTME;
    if (fabs(aUTMN - gpsc->oldutmn) >= MINUTM)   wparam |= COORD_UTMN;
    if (wparam != 0) {
      if (cbthr_enter_cs(&(gpsc->cs)) == 0) {
        if ((wparam & COORD_UTC) == COORD_UTC) {
          SPRTF(gpsc->sutc, VAL_STR_LEN, _szFmtUTC, anUTC);
          gpsc->sutc[VAL_STR_LEN] = TEXT('\0');
          gpsc->oldutc = anUTC;
        }
        if ((wparam & COORD_LAT) == COORD_LAT) {
          if (anNLat >= 0.0) 
            SPRTF(gpsc->slat, VAL_STR_LEN, _szFmtNLat, anNLat);
          else
            SPRTF(gpsc->slat, VAL_STR_LEN, _szFmtSLat, -anNLat);
          gpsc->slat[VAL_STR_LEN] = TEXT('\0');
          gpsc->oldlat = anNLat;
          dogpx = 1;
        }
        if ((wparam & COORD_LON) == COORD_LON) {
          if (anELon >= 0.0) 
            SPRTF(gpsc->slon, VAL_STR_LEN, _szFmtELon, anELon);
          else
            SPRTF(gpsc->slon, VAL_STR_LEN, _szFmtWLon, -anELon);
          gpsc->slon[VAL_STR_LEN] = TEXT('\0');
          gpsc->oldlon = anELon;
          dogpx = 1;
        }
        if ((wparam & COORD_ALT) == COORD_ALT) {
          SPRTF(gpsc->salt, VAL_STR_LEN, _szFmtAlt, anAlt);
          gpsc->salt[VAL_STR_LEN] = TEXT('\0');
          gpsc->oldalt = anAlt;
          dogpx = 1;
        }
        if ((wparam & COORD_UTMN) == COORD_UTMN) {
          SPRTF(gpsc->sutmn, VAL_STR_LEN, _szFmtUTMN, aUTMN);
          gpsc->sutmn[VAL_STR_LEN] = TEXT('\0');
          gpsc->oldutmn = aUTMN;
        }
        if ((wparam & COORD_UTME) == COORD_UTME) {
          SPRTF(gpsc->sutme, VAL_STR_LEN, _szFmtUTME, aUTME);
          gpsc->sutme[VAL_STR_LEN] = TEXT('\0');
          gpsc->oldutme = aUTME;
        }
        if (cbthr_leave_cs(&(gpsc->cs)) == 0) {
          PostMessage(gpsc->hwnd, WM_NEWCOORD, wparam, 0);
        }
        if (dogpx) cbgpx_write(&(gpsc->hgpx), anNLat, anELon, anAlt);
      }
    } /* if (wparam != 0) */
  } /* if (aCtx != NULL) */
} /* _gpscb() */

/** \brief Creates the value for enabling or disabling a menu item depending
  *        on the given flag and mask.
  * \param enmask   the bitwise mask of currently enabled menu items.
  * \param flag     eg. MENU_START_ENABLED, MENU_STOP_ENABLED etc.
  */
#define ENMENU(enmask, flag) \
 (((enmask) & (flag)) == (flag))? (MF_BYCOMMAND | MF_ENABLED) :\
                                  (MF_BYCOMMAND | MF_GRAYED)

/** \brief Shows a popup menu at the given position.
  * \param hwnd               the window, within which to display the menu.
  * \param x                  the x position
  * \param y                  the y position
  * \param enabled            bitwise mask of enabled menu items
  */
static void _popupmenu(HWND hwnd, int x, int y, int enabled) {
  TPMPARAMS tpm;
  HMENU     mmenu, menu;
  POINT     pt;

  pt.x = x; pt.y = y;
  MapWindowPoints(hwnd, HWND_DESKTOP, &pt, 1);
  tpm.cbSize = sizeof(tpm);
  GetClientRect(hwnd, &(tpm.rcExclude));
  mmenu = LoadMenu(_hresmod, MAKEINTRESOURCE(M_MAIN));
  menu = GetSubMenu(mmenu, 0);
  EnableMenuItem(mmenu, M_START, ENMENU(enabled, MENU_START_ENABLED));
  EnableMenuItem(mmenu, M_STOP, ENMENU(enabled, MENU_STOP_ENABLED));
  EnableMenuItem(mmenu, M_CFG, ENMENU(enabled, MENU_CONFIG_ENABLED));
  EnableMenuItem(mmenu, M_GPXSTART, ENMENU(enabled, MENU_GPXSTART_ENABLED));
  EnableMenuItem(mmenu, M_GPXSTOP, ENMENU(enabled, MENU_GPXSTOP_ENABLED));
  EnableMenuItem(mmenu, M_EXIT, ENMENU(enabled, MENU_EXIT_ENABLED));
  TrackPopupMenuEx(menu, TPM_LEFTALIGN | TPM_TOPALIGN, pt.x, pt.y, hwnd, &tpm);
  DestroyMenu(mmenu);
  DestroyMenu(menu);
} /* _popupmenu() */

#ifdef UNDER_CE
/** \brief Switches the PDA display off.
  *
  * http://www.pinvoke.net/default.aspx/coredll/ExtEscape.html
  */
static void _displayoff() {
  static const int  SETPOWERMANAGEMENT = 6147;
  static const BYTE VIDEOPOWEROFF = 4;
  static const BYTE DPMSVER = 1;
  HDC  hdc;
  char vpm[12];

  ZeroMemory(vpm, sizeof(vpm));
  vpm[0] = 12;
  vpm[4] = DPMSVER;
  vpm[8] = VIDEOPOWEROFF;
  hdc = GetDC(NULL);
  ExtEscape(hdc, SETPOWERMANAGEMENT, 12, vpm, 0, NULL);
} /* _displayoff() */
#endif

/** \brief Opens common dialog for selecting the gpx file to create.
  * \param hwnd         the window owning the common dialog for selecting
  *                     the file name.
  * \param theFileName  buffer where to put the selected file name.
  * \param aLen         lenght of the buffer.
  * \return !=0 if the user selected a file name, 0 else or on error.
  */
static short _dlggpxfilename(HWND hwnd, TCHAR *theFileName, size_t aLen) {
  OPENFILENAME of;
  short        retval;

  retval = 0;
  if (theFileName != NULL && aLen > 0) {
    theFileName[0] = TEXT('\0');
    ZeroMemory(&of, sizeof(of));
    of.lStructSize = sizeof(of);
    of.hwndOwner = hwnd;
    of.lpstrFile = theFileName;
    of.nMaxFile = aLen;
    of.lpstrFilter = _szGPXFilter;
    of.Flags = 0;
    if (GetSaveFileName(&of)) retval = 1;
  }
  return retval;
} /* _dlggpxfilename() */

/** \brief Configuration dialog procedure.
  *
  * Puts the presently configured values into the controls of the dialog.
  * If the dialog box has been closed using the OK button, then it puts the
  * values from the controls back into the current configuration. The current
  * configuration structure is expected under lParam.
  * \param hwnd     the handle of the configuration dialog
  * \param message  the windows message, WM_INITDIALOH, WM_COMMAND and
  *                 WM_CBSIREADY are handled in this procedure.
  * \param wParam   the procedure  expects here the ID of the button pressed
  *                 (IDOK, IDCANCEL) on WM_COMMAND.
  * \param lParam   the pointer to struct _config containing the current
  *                 configuration is expected here for WM_INITDIALOG.
  * \return TRUE for WM_INITDIALOG and WM_COMMAND, FALSE else.
  */
static BOOL CALLBACK _cfgdlgproc(HWND   hwnd,
                                 UINT   message,
                                 WPARAM wParam,
                                 LPARAM lParam) {
  static t_cbserinfo     cbsi;
  static struct _cbsictx cc = { 0 };
  static HWND            hportname, hbaud, hdatabits, hparity, hstopbits, hok;
  static struct _config* cfg = NULL;
  struct _cbsientry      *ce;
  BOOL                   retval;

  retval = FALSE;
  switch(message) {
    case WM_INITDIALOG :
      cfg = (struct _config *) lParam;
      _cbsientry_release(cc.first); cc.first = NULL;
      cc.hwnd = hwnd;
      cbserinitinfo(&cbsi);
      hportname = GetDlgItem(hwnd, CB_PORT);
      hbaud     = GetDlgItem(hwnd, CB_SPEED);
      hdatabits = GetDlgItem(hwnd, CB_DATABITS);
      hparity   = GetDlgItem(hwnd, CB_PARITY);
      hstopbits = GetDlgItem(hwnd, CB_STOPBITS);
      hok       = GetDlgItem(hwnd, IDOK);
      SendMessage(hportname, CB_RESETCONTENT, 0, 0);
      SendMessage(hbaud,     CB_RESETCONTENT, 0, 0);
      SendMessage(hdatabits, CB_RESETCONTENT, 0, 0);
      SendMessage(hparity,   CB_RESETCONTENT, 0, 0);
      SendMessage(hstopbits, CB_RESETCONTENT, 0, 0);
      EnableWindow(hportname, FALSE);
      EnableWindow(hbaud, FALSE);
      EnableWindow(hdatabits, FALSE);
      EnableWindow(hparity, FALSE);
      EnableWindow(hstopbits, FALSE);
      EnableWindow(hok, FALSE);
      cbserinfo(&cbsi, _cbserinfocb, &cc);
      retval = TRUE;
      break;
    case WM_COMMAND :
      switch(LOWORD(wParam)) {
        case IDOK :
          retval = TRUE;
          cbserinfo_end(&cbsi, 1);
          SendMessage(hportname, WM_GETTEXT,
                      CFG_STR_LEN,
                      (LPARAM) (cfg->portname));
          SendMessage(hbaud, WM_GETTEXT,
                      CFG_STR_LEN,
                      (LPARAM) (cfg->baudrate));
          SendMessage(hdatabits, WM_GETTEXT,
                      CFG_STR_LEN,
                      (LPARAM) (cfg->databits));
          SendMessage(hparity, WM_GETTEXT,
                      CFG_STR_LEN,
                      (LPARAM) (cfg->parity));
          SendMessage(hstopbits, WM_GETTEXT,
                      CFG_STR_LEN,
                      (LPARAM) (cfg->stopbits));
          cfg->portid = (short) SendMessage(hportname, CB_GETCURSEL, 0, 0);
          cfg->baudid = (short) SendMessage(hbaud, CB_GETCURSEL, 0, 0);
          cfg->databitsid = (short) SendMessage(hdatabits, CB_GETCURSEL, 0, 0);
          cfg->parityid = (short) SendMessage(hparity, CB_GETCURSEL, 0, 0);
          cfg->stopbitsid = (short) SendMessage(hstopbits, CB_GETCURSEL, 0, 0);
          EndDialog(hwnd, 1);
          break;
        case IDCANCEL :
          retval = TRUE;
          cbserinfo_end(&cbsi, 1);
          _cbsientry_release(cc.first); cc.first = NULL;
          EndDialog(hwnd, 0);
          break;
        default :
          break;
      }
      break;
    case WM_CBSIREADY :
      for(ce = cc.first; ce != NULL; ce = ce->next) {
        switch(ce->typ) {
          case CBSI_SERIALPORT :
            SendMessage(hportname, CB_ADDSTRING, 0, (LPARAM) (ce->val));
            break;
          case CBSI_BAUDRATE :
            SendMessage(hbaud, CB_ADDSTRING, 0, (LPARAM) (ce->val));
            break;
          case CBSI_DATABITS :
            SendMessage(hdatabits, CB_ADDSTRING, 0, (LPARAM) (ce->val));
            break;
          case CBSI_PARITY :
            SendMessage(hparity, CB_ADDSTRING, 0, (LPARAM) (ce->val));
            break;
          case CBSI_STOPBITS :
            SendMessage(hstopbits, CB_ADDSTRING, 0, (LPARAM) (ce->val));
            break;
          default :
            break;
        } /* switch */
      } /* for ce */
      EnableWindow(hportname, TRUE);
      EnableWindow(hbaud, TRUE);
      EnableWindow(hdatabits, TRUE);
      EnableWindow(hparity, TRUE);
      EnableWindow(hstopbits, TRUE);
      EnableWindow(hok, TRUE);
      if (cfg != NULL) {
        if (cfg->portid >= 0)
          SendMessage(hportname, CB_SETCURSEL, cfg->portid, 0);
        if (cfg->baudid >= 0)
          SendMessage(hbaud, CB_SETCURSEL, cfg->baudid, 0);
        if (cfg->databitsid >= 0)
          SendMessage(hdatabits, CB_SETCURSEL, cfg->databitsid, 0);
        if (cfg->parityid >= 0)
          SendMessage(hparity, CB_SETCURSEL, cfg->parityid, 0);
        if (cfg->stopbitsid >= 0)
          SendMessage(hstopbits, CB_SETCURSEL, cfg->stopbitsid, 0);
      }
      cbserinfo_end(&cbsi, 1);
      _cbsientry_release(cc.first); cc.first = NULL;
      retval = TRUE;
      break;
    default :
      break;
  } /* switch(message) */
  return retval;
} /* _cfgdlgproc() */

/** \brief Builds the configuration file name.
  * The configuration file name is [ApplicationData]\[_APPNAME].cfg, if not compiled
  * with -DUSE_ENV, else it is $HOME/[_APPNAME].cfg
  * \param theFileName buffer for the resulting file name, must be able to
  *                    hold at least MAX_PATH+1 characters.
  */
static void _cfgfilename(TCHAR *theFileName) {
#ifdef __unix
  /* this is the case, if compiling with unix and wine */
# ifdef UNICODE
  static const TCHAR *DIRSEP = L"\\";
# else
  static const TCHAR *DIRSEP = "\\";
# endif
#else
# ifdef UNICODE
  static const TCHAR *DIRSEP = L"/";
# else
  static const TCHAR *DIRSEP = "/";
# endif
#endif
  size_t     lenapp, lendir;

  if (theFileName != NULL) {
#ifdef USE_ENV
    {
      const char *homedir;

      homedir = getenv("HOME");
      if (homedir == NULL) homedir = getenv("USERPROFILE");
      if (homedir == NULL) homedir = ".";
# ifdef UNICODE
        mbstowcs(theFileName, homedir, MAX_PATH);
# else
        strncpy(theFileName, homedir, MAX_PATH);
# endif
    }
#else
    SHGetSpecialFolderPath(0, theFileName, CSIDL_APPDATA, TRUE);
#endif
#ifdef UNICODE
    lendir = wcslen(theFileName);
    lenapp = wcslen(_APPNAME);
#else
    lendir = strlen(theFileName);
    lenapp = strlen(_APPNAME);
#endif
    if (lendir + lenapp + 5 < MAX_PATH) {
#ifdef UNICODE
      wcscat(theFileName, DIRSEP);
      wcscat(theFileName, _APPNAME);
      wcscat(theFileName, L".cfg");
#else
      strcat(theFileName, DIRSEP);
      strcat(theFileName, _APPNAME);
      strcat(theFileName, ".cfg");
#endif
    }
  }
} /* _cfgfilename() */

/** \brief Reads configuration from a file into the given config structure.
  *
  * If the file doesn't exist or if reading the configuration fails for some
  * reason, then the configuration structure is filled with plausible defaults.
  * \param aCfg  pointer to the configuration structure.
  * \sa _cfgfilename()
 */
static void _rdcfg(struct _config *aCfg) {
  TCHAR  cfgfile[MAX_PATH+1];
  BYTE   *ptr;
  HANDLE hf;
  DWORD  remaining, len;
  short  isok;

  if (aCfg != NULL) {
    isok = 0;
    remaining = (DWORD) sizeof(struct _config);
    _cfgfilename(cfgfile);
    hf = CreateFile(cfgfile,
                    GENERIC_READ,
                    0,
                    NULL,
                    OPEN_EXISTING,
                    FILE_ATTRIBUTE_NORMAL,
                    NULL);
    if (hf != INVALID_HANDLE_VALUE) {
      ptr = (BYTE *) aCfg;
      while(remaining > 0 &&
            ReadFile(hf, ptr, remaining, &len, NULL) != 0 &&
            len > 0) {
        remaining -= len;
        ptr += len;
      }
      CloseHandle(hf);
      if (remaining == 0) isok = 1;
    }
    if (isok == 0) {
      /* configuration not read properly, fill it with defaults */
      aCfg->portid = 0;
      aCfg->baudid = 0;
      aCfg->databitsid = 0;
      aCfg->parityid = 0;
      aCfg->stopbitsid = 0;
#ifdef UNICODE
      mbstowcs(aCfg->portname, DEFAULTPORT, CFG_STR_LEN);
      mbstowcs(aCfg->baudrate, DEFAULTSPEED, CFG_STR_LEN);
      mbstowcs(aCfg->databits, DEFAULTDATABITS, CFG_STR_LEN);
      mbstowcs(aCfg->parity, DEFAULTPARITY, CFG_STR_LEN);
      mbstowcs(aCfg->stopbits, DEFAULTSTOPBITS, CFG_STR_LEN);
#else
      strncpy(aCfg->portname, DEFAULTPORT, CFG_STR_LEN);
      strncpy(aCfg->baudrate, DEFAULTSPEED, CFG_STR_LEN);
      strncpy(aCfg->databits, DEFAULTDATABITS, CFG_STR_LEN);
      strncpy(aCfg->parity, DEFAULTPARITY, CFG_STR_LEN);
      strncpy(aCfg->stopbits, DEFAULTSTOPBITS, CFG_STR_LEN);
#endif
      aCfg->portname[CFG_STR_LEN] = TEXT('\0');
      aCfg->baudrate[CFG_STR_LEN] = TEXT('\0');
      aCfg->databits[CFG_STR_LEN] = TEXT('\0');
      aCfg->parity[CFG_STR_LEN] = TEXT('\0');
      aCfg->stopbits[CFG_STR_LEN] = TEXT('\0');
    }
  } /* if (aCfg != NULL) */
} /* _rdcfg() */

/** \brief Writes configuration to the configuration file.
  *
  * The configuration file name is always [ApplicationData]\[_APPNAME].cfg.
  * \param aCfg  pointer to the configuration structure.
  * \sa _cfgfilename()
  */
static void _wrcfg(const struct _config *aCfg) {
  TCHAR       cfgfile[MAX_PATH+1];
  const BYTE* ptr;
  HANDLE      hf;
  DWORD       remaining, len;

  if (aCfg != NULL) {
    remaining = (DWORD) sizeof(struct _config);
    _cfgfilename(cfgfile);
    hf = CreateFile(cfgfile,
                    GENERIC_WRITE,
                    FILE_SHARE_READ,
                    NULL,
                    CREATE_ALWAYS,
                    FILE_ATTRIBUTE_NORMAL,
                    NULL);
    if (hf != INVALID_HANDLE_VALUE) {
      ptr = (const BYTE *) aCfg;
      while(WriteFile(hf, ptr, remaining, &len, NULL) != 0 && remaining > 0) {
        remaining -= len;
        ptr += len;
      }
      CloseHandle(hf);
    }
  } /* if (aCfg != NULL) */
} /* _wrcfg() */

/** \brief The standard MS Windows window procedure.
  *
  * On WM_INITDIALOG the function initialozes internal data structure:
  * it reads the current configuration from the configuration file into
  * a variable of type struct _config, initializes the GPS handle etc.
  * On WM_COMMAND it handles the menu items pressed: starts the GPS reading
  * thread on M_START, stops it on M_STOP, starts the configuration
  * dialog on M_CFG or exits the application on M_EXIT.
  * \param hwnd    the handle of the window
  * \param message window message, WM_INITDIALOG, WM_COMMAND, WM_CLOSE,
  *                WM_LBUTTONDOWN, WM_RBUTTONDOWN.
  * \param wParam  message parameter, on WM_COMMAND M_START, M_STOP,
  *                M_CFG or M_EXIT are handled here.
  * \param lParam  message parameter, ignored.
  * \return 0, if the procedure handles the given message by itself,
  *         return value of DefWindowProc() else
  */
static LRESULT CALLBACK _wndproc(HWND   hwnd,
                                 UINT   message,
                                 WPARAM wParam,
                                 LPARAM lParam) {
#ifdef UNDER_CE
  static const UINT            TIMERID = 1;
#endif
  static struct _gpscbcontrols gpscbctx;
  static t_cbgps               hgps;
  static struct _config        cfg;
  static int                   enabled_menu_items = MENU_ALL_ENABLED;
  int                          retval, res;

  switch(message) {
    case WM_INITDIALOG :
      ZeroMemory(&gpscbctx, sizeof(gpscbctx));
      cbthr_init_cs(&(gpscbctx.cs));
      gpscbctx.oldutc = 0.0;
      gpscbctx.oldlat = 0.0;
      gpscbctx.oldlon = 0.0;
      gpscbctx.oldalt = 0.0;
      gpscbctx.oldutme = 0.0;
      gpscbctx.oldutmn = 0.0;
      gpscbctx.hwnd = hwnd;
      enabled_menu_items = MENU_ALL_ENABLED & 
                           ~MENU_STOP_ENABLED & ~MENU_GPXSTOP_ENABLED;
      cbgpsinit(&hgps);
      _rdcfg(&cfg);
      retval = 0;
      break;
    case WM_COMMAND :
      switch(LOWORD(wParam)) {
        case M_START :
          res = cbgpsclose(&hgps);
          if (res != 0) {
            res =
#ifdef UNICODE
                  cbgpsopenwstr
#else
                  cbgpsopenstr
#endif
                              (&hgps,
                               cfg.portname,
                               cfg.baudrate,
                               cfg.databits,
                               cfg.parity,
                               cfg.stopbits,
                               DEFAULT_TIMEOUT,
                               DEFAULT_SYSTEM,
                               _gpscb,
                               &gpscbctx);
            if (res != 0) {
              enabled_menu_items &= ~MENU_START_ENABLED;
              enabled_menu_items |= MENU_STOP_ENABLED;
#ifdef UNDER_CE
              SetTimer(hwnd, TIMERID, 1000, NULL);
#endif
            }
            else {
              MessageBox(hwnd, _szOpenErr, _APPNAME, MB_ICONERROR);
            }
          }
          else {
            MessageBox(hwnd, _szCloseErr, _APPNAME, MB_ICONERROR);
          }
          break;
        case M_STOP :
          res = cbgpsclose(&hgps);
          if (res != 0) {
            enabled_menu_items &= ~MENU_STOP_ENABLED;
            enabled_menu_items |= MENU_START_ENABLED;
#ifdef UNDER_CE
            KillTimer(hwnd, TIMERID);
#endif
          }
          else {
            MessageBox(hwnd, _szCloseErr, _APPNAME, MB_ICONERROR);
          }
          break;
        case M_CFG :
          DialogBoxParam(_hresmod,
                         MAKEINTRESOURCE(DLG_CBCFG),
                         hwnd,
                         _cfgdlgproc,
                         (LPARAM) (&cfg));
          break;
#ifdef UNDER_CE
        case M_DOFF :
          _displayoff();
          break;
#endif
        case M_GPXSTART :
          {
            TCHAR filename[MAX_PATH];

            if (_dlggpxfilename(hwnd, filename, MAX_PATH) != 0) {
              cbgpx_close(&(gpscbctx.hgpx));
#ifdef UNICODE
              res = cbgpx_openw(&(gpscbctx.hgpx), filename);
#else
              res = cbgpx_opena(&(gpscbctx.hgpx), filename);
#endif
              if (res != 0) {
                enabled_menu_items &= ~MENU_GPXSTART_ENABLED;
                enabled_menu_items |= MENU_GPXSTOP_ENABLED;
              }
            }
          }
          break;
        case M_GPXSTOP :
          cbgpx_close(&(gpscbctx.hgpx));
          enabled_menu_items &= ~MENU_GPXSTOP_ENABLED;
          enabled_menu_items |= MENU_GPXSTART_ENABLED;
          break;
        case M_EXIT :
          SendMessage(hwnd, WM_CLOSE, 0, 0);
          break;
        default :
          break;
      } /* switch */
      retval = 0;
      break;
    case WM_CLOSE :
      retval = 0;
      if (MessageBox(hwnd, _szQuit, _APPNAME, MB_YESNO) == IDYES) {
        cbgpx_close(&(gpscbctx.hgpx));
        cbgpsclose(&hgps);
        _wrcfg(&cfg);
#ifdef UNDER_CE
        KillTimer(hwnd, TIMERID);
#endif
        PostQuitMessage(0);
      }
      break;
#ifdef UNDER_CE
    case WM_TIMER :
      retval = 0;
      SystemIdleTimerReset();
      break;
#endif
    case WM_LBUTTONDOWN :
    case WM_RBUTTONDOWN :
      _popupmenu(hwnd, LOWORD(lParam), HIWORD(lParam), enabled_menu_items);
      break;
    case WM_NEWCOORD :
      if (wParam != 0) {
        if (cbthr_enter_cs(&(gpscbctx.cs)) == 0) {
          if ((wParam & COORD_UTC) == COORD_UTC)
            SetDlgItemText(hwnd, T_UTC, gpscbctx.sutc);
          if ((wParam & COORD_ALT) == COORD_ALT)
            SetDlgItemText(hwnd, T_ALT, gpscbctx.salt);
          if ((wParam & COORD_LAT) == COORD_LAT)
            SetDlgItemText(hwnd, T_LAT, gpscbctx.slat);
          if ((wParam & COORD_LON) == COORD_LON)
            SetDlgItemText(hwnd, T_LON, gpscbctx.slon);
          if ((wParam & COORD_UTMN) == COORD_UTMN)
            SetDlgItemText(hwnd, T_UTMN, gpscbctx.sutmn);
          if ((wParam & COORD_UTME) == COORD_UTME)
            SetDlgItemText(hwnd, T_UTME, gpscbctx.sutme);
          cbthr_leave_cs(&(gpscbctx.cs));
        }
      }
      break;
    default :
      retval = DefWindowProc(hwnd, message, wParam, lParam);
      break;
  } /* switch */
  return retval;
} /* _wndproc() */

#ifndef __WINE__
/** \brief Returns module handle for the resource dll responsible for the
  *        current locale.
  *
  * The function looks for resource DLLs having the directory of the
  * current executable plus filename of the executable plus a 4 digit
  * hex number ID of the language +
  * extension dll. The language ID according to the return value of
  * GetUserDefaultLangID or GetSystemDefaultLangID (eg. 0407 hex for DE,
  * 0409 hex for EN, 0415 hex for PL). If the DLL with the appropriate
  * name has been found, the the function loads this DLL using LoadLibrary
  * and returns the handle of it. If no suitable DLLs have been found for
  * the user or system default laungage ID, then GetModuleHandle(NULL)
  * (the handle of the exe) is returned meaning, that the application has
  * to use the default locales.
  *
  * Example of the dll file name for the locale:
  * -# current executable located in c:\/emp/cb.exe
  * -# name of the locale resource DLL for Polish c:/temp/cb0415.dll
  * -# name of the locale resource DLL for German c:/temp/cb0407.dll
  *
  * Wondering, why not for wine? Wine doesn't need this complicated
  * way of localized resource dll loading. We can put all localized
  * resources into one rc file and wine will automatically select
  * the proper resource for the current system language in the
  * environment variable LANG. This is because
  * wine properly sets GetUser/SystemDefaultUILanguage() and
  * windows doesn't. The default UI language in windows is this
  * one, which you buyed in windows: spanish windows - spanish UI.
  *
  * \param defHandle the default handle to return, if the DLLs have not
  *                  been found.
  * \return the handle of the module for resources to use with current
  *         locales.
  */
static HMODULE _nlsresmod(HMODULE defHandle) {
  TCHAR   prefix[MAX_PATH+1], dllfile[MAX_PATH+1];
  TCHAR   *ptr, *ptrprev;
  LANGID  langs[2];
  HMODULE retval, hm;
  int     i, len;

  langs[0] = GetUserDefaultLangID();
  langs[1] = GetSystemDefaultLangID();
  retval = defHandle;
  len = GetModuleFileName(NULL, prefix, MAX_PATH);
  if (len > 0) {
    prefix[MAX_PATH] = TEXT('\0');
    ptrprev = NULL;
    ptr = prefix;
#ifdef UNICODE
    while((ptr = wcsstr(ptr, TEXT(".exe"))) != NULL)
#else
    while((ptr = strstr(ptr, TEXT(".exe"))) != NULL)
#endif
    {
      ptrprev = ptr;
      ptr++;
    }
    if (ptrprev != NULL) *ptrprev = TEXT('\0');
    for(i = 0; i < 2; i++) {
#ifdef UNICODE
      SPRTF(dllfile, MAX_PATH, TEXT("%ls%04X.dll"), prefix, langs[i]);
#else
      SPRTF(dllfile, MAX_PATH, TEXT("%s%04X.dll"), prefix, langs[i]);
#endif
      dllfile[MAX_PATH] = TEXT('\0');
      hm = LoadLibrary(dllfile);
      if (hm != NULL) {
        retval = hm;
        break;
      }
    } /* for i */
  }
  return retval;
} /* _nlsresmod() */
#endif

/** \brief The main function of the application.
  *
  * Sets the dialog cbmain as the main window and _wndproc as the window procedure.
  *
  * \param hInstance     the handle of the instance.
  * \param hPrevInstance ignored.
  * \param szCmdLine     ignored.
  * \param iCmdShow      needed for ShowWindow.
  */
int WINAPI WinMain(HINSTANCE hInstance,
                   HINSTANCE hPrevInstance,
#ifdef UNDER_CE
                   LPWSTR    szCmdLine,
#else
                   PSTR      szCmdLine,
#endif
                   int       iCmdShow) {
  WNDCLASS wndclass;
  MSG      msg;
  HWND     hwnd;
  int      retval;

  retval = 0;
  hwnd = FindWindow(_APPNAME, NULL);
  if (hwnd == 0) {
    /* run it as usual, if not already running */
#ifdef __WINE__
    _hresmod = hInstance;
#else
    _hresmod = _nlsresmod(hInstance);
#endif
    wndclass.style         = CS_HREDRAW | CS_VREDRAW;
    wndclass.lpfnWndProc   = _wndproc;
    wndclass.cbClsExtra    = 0;
#ifdef __CEGCC__
    wndclass.cbWndExtra    = 0;
#else
    wndclass.cbWndExtra    = DLGWINDOWEXTRA;
#endif
    wndclass.hInstance     = _hresmod;
    wndclass.hIcon         = LoadIcon(hInstance, MAKEINTRESOURCE(MAINICON));
    wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wndclass.hbrBackground = (HBRUSH) (COLOR_WINDOW+1);
    wndclass.lpszMenuName  = NULL;
    wndclass.lpszClassName = _APPNAME;
    if (RegisterClass(&wndclass) != 0 &&
        LoadString(_hresmod, S_QUIT, _szQuit, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_CLOSEERR, _szCloseErr, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_OPENERR, _szOpenErr, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_UTCFMT, _szFmtUTC, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_NLATFMT, _szFmtNLat, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_SLATFMT, _szFmtSLat, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_ELONFMT, _szFmtELon, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_WLONFMT, _szFmtWLon, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_ALTFMT, _szFmtAlt, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_UTMEFMT, _szFmtUTME, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_UTMNFMT, _szFmtUTMN, MAX_STR_LEN) > 0 &&
        LoadString(_hresmod, S_GPXFILTER, _szGPXFilter, MAX_STR_LEN) > 0) {
      hwnd = CreateDialog(_hresmod,
                          MAKEINTRESOURCE(DLG_CB),
                          0,
                          (DLGPROC) _wndproc);
      ShowWindow(hwnd, iCmdShow);
      UpdateWindow(hwnd);
      while(GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
      if (_hresmod != hInstance) FreeLibrary(_hresmod);
      retval = msg.wParam;
    }
    else {
      _fmtErrMsg();
    }
  } /* if (hwnd == 0) */
  else {
    /* activate the window of the running app, if already running */
    SetForegroundWindow((HWND)(((DWORD) hwnd) | 0x01));
  }
  return retval;
} /* WinMain() */

#ifdef OWN_STARTUP
/** \brief Main entry point for applications build for pocket PC on
  *        MS Visual Studio 8 without oldnames.lib, which is not available
  *        for this platform.
  *
  * See also http://www.hailstorm.net/papers/smallwin32.htm
  */
int WINAPI WinMainCRTStartup(void) {
  int             retval;

  retval = WinMain(GetModuleHandle(NULL), NULL, GetCommandLine(), SW_SHOW);
  return retval;
} /* WinMainCRTStartup() */
#endif

