#include "cbser.h"
#include <stdlib.h>
#ifndef FOR_MS_WINDOWS
# include <string.h>
# include <unistd.h>
# include <termios.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <dirent.h>
# include <fnmatch.h>
# include <stdio.h>
#endif

/** \brief Maximal number of bytes to read from the com port at once.
  */
#define BUFSIZE     1023

#ifdef FOR_MS_WINDOWS

# ifdef UNICODE
#  define _openport(comName, res) { \
    wchar_t wComName[CBSTXTLEN+1]; \
    mbstowcs(wComName, comName, CBSTXTLEN); \
    wComName[CBSTXTLEN] = L'\0'; \
    res = CreateFile(wComName,GENERIC_READ|GENERIC_WRITE,0,0,OPEN_EXISTING,0,0); \
}
# else
#  define _openport(comName, res) { \
    res = CreateFile(comName,GENERIC_READ|GENERIC_WRITE,0,0,OPEN_EXISTING,0,0); \
}
# endif
# define _closeport(comHandle) CloseHandle(comHandle);

typedef DWORD _t_int;

const t_cbserparam COMSPEEDS[] = {
    { "256000", CBR_256000 },
    { "128000", CBR_128000 },
    { "115200", CBR_115200 },
    {  "57600",  CBR_57600 },
    {  "38400",  CBR_38400 },
    {  "19200",  CBR_19200 },
    {   "9600",   CBR_9600 },
    {   "4800",   CBR_4800 },
    {   "2400",   CBR_2400 },
    {   "1200",   CBR_1200 },
    {    "600",    CBR_600 },
    {    "300",    CBR_300 },
    {    "110",    CBR_110 }
  };

const t_cbserparam COMDATABITS[] = {
     {  "8", 8 },
     {  "7", 7 },
     {  "6", 6 },
     {  "5", 5 }
  };

const t_cbserparam COMPARITIES[] = {
     { "E", EVENPARITY },
     { "N", NOPARITY },
     { "O", ODDPARITY }
  };

const t_cbserparam COMSTOPBITS[] = {
     {   "1", ONESTOPBIT },
     {   "2", TWOSTOPBITS }
  };

const t_hcomport CBINVALIDCOMPORT = INVALID_HANDLE_VALUE;

/** \brief Possibly available comports on the windows system.
  * \sa cbserinfo()
  */
static const char* _COMPORTS[] = {
    "COM0:", "COM1:", "COM2:", "COM3:", "COM4:",
    "COM5:", "COM6:", "COM7:", "COM8:", "COM9:"
  };

/** \brief Number of COM port names in _COMPORTS.
  */
static const int _NCOMPORTS = sizeof(_COMPORTS)/sizeof(char*);

const char *DEFAULTPORT = "COM1:";

#else

# define _openport(comName, res) { \
  res = open(comName, O_RDWR | O_NOCTTY | O_NDELAY); \
}
# define _closeport(comHandle) close(comHandle);

typedef int _t_int;

const t_cbserparam COMSPEEDS[] = {
    { "460800", B460800 },
    { "230400", B230400 },
    { "115200", B115200 },
    {  "57600",  B57600 },
    {  "38400",  B38400 },
    {  "19200",  B19200 },
    {   "9600",   B9600 },
    {   "4800",   B4800 },
    {   "2400",   B2400 },
    {   "1200",   B1200 },
    {    "600",    B600 },
    {    "300",    B300 },
    {    "110",    B110 },
    {     "75",     B75 },
    {     "50",     B50 }
  };

const t_cbserparam COMDATABITS[] = {
     {  "8", CS8 },
     {  "7", CS7 },
     {  "6", CS6 },
     {  "5", CS5 }
  };

const t_cbserparam COMPARITIES[] = {
     { "E", PARENB },
     { "N",      0 },
     { "O", PARODD | PARENB }
  };

const t_cbserparam COMSTOPBITS[] = {
     {   "1",      0 },
     {   "2", CSTOPB }
  };

const t_hcomport CBINVALIDCOMPORT = -1;

/** \brief Directory containing device files on the unix system.
  */
static const char *_DEVDIR = "/dev/";

/** \brief Regular expressions for device files in _DEVDIR for possibly
  *        available comports.
  * \sa cbserinfo()
  */
static const char* _COMPORTS[] = {
    "ttyS*", "ttyUSB*", "cua*", "tty[a-z]"
  };

/** \brief Number of COM port names in _COMPORTS.
  */
static const int _NCOMPORTS = sizeof(_COMPORTS)/sizeof(char*);

#if defined(linux) || defined(__linux)
const char *DEFAULTPORT = "/dev/ttyS0";
#else
#if defined(sun)
const char *DEFAULTPORT = "/dev/ttya";
#else
const char *DEFAULTPORT = NULL;
#endif
#endif

#endif /* #ifdef FOR_MS_WINDOWS */

const char *DEFAULTSPEED = "9600";
const char *DEFAULTDATABITS = "8";
const char *DEFAULTPARITY = "N";
const char *DEFAULTSTOPBITS = "1";

const int NCOMSPEEDS = sizeof(COMSPEEDS) / sizeof(t_cbserparam);
const int NCOMDATABITS = sizeof(COMDATABITS) / sizeof(t_cbserparam);
const int NCOMPARITIES = sizeof(COMPARITIES) / sizeof(t_cbserparam);
const int NCOMSTOPBITS = sizeof(COMSTOPBITS) / sizeof(t_cbserparam);

void cbserinitcom(t_comport *theHandle) {
  if (theHandle != NULL) {
    theHandle->goon = 0;
    theHandle->rdCb = NULL;
    theHandle->ctx = NULL;
    theHandle->hthr = CBINVALIDTHREAD;
    theHandle->hcomport = CBINVALIDCOMPORT;
  }
} /* cbserinitcom() */

void cbserinitinfo(t_cbserinfo *theHandle) {
  if (theHandle != NULL) {
    theHandle->goon = 0;
    theHandle->cbsicb = NULL;
    theHandle->hthr = CBINVALIDTHREAD;
  }
} /* cbserinitinfo() */

/** \brief Serial port reading thread.
  *
  * Reads data from serial ports and splits it into lines of text.
  * The lines of text are assumed to be separated by carriage return
  * or newline by default. For each line of text the callback function
  * defined in the given params is called. The reading operation stops,
  * if the field goon in the given params becomes 0.
  * \param params  casted to the pointer to t_comport.
  * \sa t_comport
  */
static CBTHRFUNCTYPE _rdthrfunc(void *params) {
  char            buf[BUFSIZE+1], line[BUFSIZE+1];
  const char      *pbuf;
  char            *pline;
  const t_comport *p;
  _t_int          bytesread, i;
  size_t          j;

  if (params != NULL) {
    p = (const t_comport *) params;
    pline = line;
    j = 0;
    while(p->goon != 0) {
#ifdef FOR_MS_WINDOWS
      if (ReadFile(p->hcomport, buf, BUFSIZE, &bytesread, NULL) == 0)
        bytesread = 0;
#else
      bytesread = read(p->hcomport, buf, BUFSIZE);
#endif
      for(pbuf = buf, i = 0; i < bytesread && p->goon != 0; i++, pbuf++) {
        if (*pbuf == '\r' || *pbuf == '\n' || j >= BUFSIZE) {
          if (j > 0) {
            *pline = '\0';
            if (p->rdCb != NULL && p->goon != 0) p->rdCb(line, j, p->ctx);
            j = 0;
            pline = line;
          }
        }
        else {
          *pline = *pbuf;
          pline++;
          j++;
        }
      } /* for */
    } /* while */
  } /* if (params != NULL) */
#ifdef FOR_MS_WINDOWS
  return 0;
#else
  return NULL;
#endif
} /* _rdthrfunc() */

/** \brief Thread function for retrieving informations about serial ports.
  * \param params casted to t_cbserinfo*
  * \sa cbserinfo()
  */
static CBTHRFUNCTYPE _cbsithrfunc(void *params) {
  const t_cbserparam *pc;
  const char         **ps;
  const t_cbserinfo  *p;
  t_hcomport         hcom;
  int                i;
#ifdef FOR_MS_WINDOWS
  const char         *portname;
#else
  char               devname[FILENAME_MAX+1];
  struct dirent      *de;
  DIR                *d;
#endif

  if (params != NULL) {
    p = (const t_cbserinfo *) params;
    for(i = 0, pc = COMSPEEDS; p->goon != 0 && i < NCOMSPEEDS; i++, pc++) {
      if (p->cbsicb != NULL) p->cbsicb(pc->cbtext, CBSI_BAUDRATE, p->ctx);
    }
    for(i = 0, pc = COMDATABITS; p->goon != 0 && i < NCOMDATABITS; i++, pc++) {
      if (p->cbsicb != NULL) p->cbsicb(pc->cbtext, CBSI_DATABITS, p->ctx);
    }
    for(i = 0, pc = COMPARITIES; p->goon != 0 && i < NCOMPARITIES; i++, pc++) {
      if (p->cbsicb != NULL) p->cbsicb(pc->cbtext, CBSI_PARITY, p->ctx);
    }
    for(i = 0, pc = COMSTOPBITS; p->goon != 0 && i < NCOMSTOPBITS; i++, pc++) {
      if (p->cbsicb != NULL) p->cbsicb(pc->cbtext, CBSI_STOPBITS, p->ctx);
    }
#ifdef FOR_MS_WINDOWS
    for(i = 0, ps = _COMPORTS; p->goon != 0 && i < _NCOMPORTS; i++, ps++) {
      portname = *ps;
       _openport(portname, hcom);
      if (hcom != CBINVALIDCOMPORT) {
        if (p->cbsicb != NULL) p->cbsicb(portname, CBSI_SERIALPORT, p->ctx);
        _closeport(hcom);
      }
    }
#else
    de = NULL;
    d = opendir(_DEVDIR);
    if (d != NULL) {
      while((de = readdir(d)) != NULL) {
        for(i = 0, ps = _COMPORTS; i < _NCOMPORTS; i++, ps++) {
#ifdef _DIRENT_HAVE_D_TYPE
          if (de->d_type == DT_CHR || de->d_type == DT_LNK) {
#else
          {
#endif
            /* we are to lazy for following links in cheking the
               type of the linked file */
            if (fnmatch(*ps, de->d_name, FNM_PATHNAME) == 0) {
              snprintf(devname, FILENAME_MAX, "%s%s", _DEVDIR, de->d_name);
              devname[FILENAME_MAX] = '\0';
              _openport(devname, hcom);
              if (hcom != CBINVALIDCOMPORT) {
                if (p->cbsicb != NULL) p->cbsicb(devname, CBSI_SERIALPORT, p->ctx);
                _closeport(hcom);
              }
            }
          }
        } /* for */
      } /* while */
    } /* if (d != NULL) */
#endif
    if (p->goon && p->cbsicb != NULL) p->cbsicb(NULL, CBSI_READY, p->ctx);
  }
#ifdef FOR_MS_WINDOWS
  return 0;
#else
  return NULL;
#endif
} /* _cbsithrfunc() */

int cbseropen(t_comport          *theHandle,
              const char         *aComPort,
              const t_cbserparam *aSpeed,
              const t_cbserparam *databits,
              const t_cbserparam *aParity,
              const t_cbserparam *stopbits,
              int                aTimeout,
              t_cbsercb          rdCb,
              void               *aCtx) {
  int retval;

  retval = 0;
  cbserinitcom(theHandle);
  if (theHandle != NULL && aComPort != NULL && aSpeed != NULL &&
      databits != NULL && aParity != NULL && stopbits != NULL) {
    theHandle->ctx = aCtx;
    _openport(aComPort, theHandle->hcomport);
    if (theHandle->hcomport != CBINVALIDCOMPORT) {
#ifdef FOR_MS_WINDOWS
      COMMTIMEOUTS cmt;
      DCB          dcb;
      DWORD        c, to;

      to = aTimeout <= 0? 1 : aTimeout;
      c = to / BUFSIZE; if (c <= 0) c = 1;
      ZeroMemory(&dcb, sizeof(dcb));
      if (!GetCommState(theHandle->hcomport, &dcb)) goto err;
      dcb.BaudRate = aSpeed->setting;
      dcb.ByteSize = (BYTE) databits->setting;
      dcb.Parity = (BYTE) aParity->setting;
      dcb.StopBits = (BYTE) stopbits->setting;
      if (!SetCommState(theHandle->hcomport, &dcb)) goto err;
      ZeroMemory(&cmt, sizeof(cmt));
      cmt.ReadIntervalTimeout = to;
      cmt.ReadTotalTimeoutMultiplier = c;
      cmt.ReadTotalTimeoutConstant = to;
      cmt.WriteTotalTimeoutMultiplier = c;
      cmt.WriteTotalTimeoutConstant = to;
      if (!SetCommTimeouts(theHandle->hcomport, &cmt)) goto err;
#else
      struct termios dcb;
      int            to;

      to = aTimeout <= 0? 1 : aTimeout;
      memset(&dcb, 0, sizeof(struct termios));
      if (tcgetattr(theHandle->hcomport, &dcb) != 0) goto err;
      if (cfsetispeed(&dcb, aSpeed->setting) != 0) goto err;
      if (cfsetospeed(&dcb, aSpeed->setting) != 0) goto err;
      dcb.c_cflag &= ~CSIZE;  dcb.c_cflag |= databits->setting;
      dcb.c_cflag &= ~PARENB; dcb.c_cflag |= aParity->setting;
      dcb.c_cflag &= ~CSTOPB; dcb.c_cflag |= stopbits->setting;
      dcb.c_cc[VMIN] = 1;
      dcb.c_cc[VTIME] = to;
      dcb.c_cflag |= (CLOCAL | CREAD);
      if (tcsetattr(theHandle->hcomport, TCSANOW, &dcb) != 0) goto err;
#endif
      theHandle->goon = 1;
      theHandle->rdCb = rdCb;
      if (cbthr_create(&(theHandle->hthr), _rdthrfunc, theHandle) == 0) {
        theHandle->hthr = CBINVALIDTHREAD;
        goto err;
      }
      retval = 1;
    }
  } /* if */
  return retval;
  err:
  retval = 0;
  cbserclose(theHandle);
  return retval;
} /* cbseropen() */

int cbseropenstr(t_comport  *theHandle,
                 const char *aComPort,
                 const char *aSpeed,
                 const char *databits,
                 const char *aParity,
                 const char *stopbits,
                 int        aTimeout,
                 t_cbsercb  rdCb,
                 void       *aCtx) {
  const t_cbserparam *bd, *db, *par, *sb, *p;
  int                retval, i;

  retval = 0;
  cbserinitcom(theHandle);
  if (aSpeed != NULL && databits != NULL &&
      aParity != NULL && stopbits != NULL) {
    bd = NULL;
    db = NULL;
    par = NULL;
    sb = NULL;
    for(i = 0, p = COMSPEEDS; i < NCOMSPEEDS; i++, p++) {
      if (strcmp(p->cbtext, aSpeed) == 0) {
        bd = p;
        break;
      }
    }
    for(i = 0, p = COMDATABITS; i < NCOMDATABITS; i++, p++) {
      if (strcmp(p->cbtext, databits) == 0) {
        db = p;
        break;
      }
    }
    for(i = 0, p = COMPARITIES; i < NCOMPARITIES; i++, p++) {
      if (strcmp(p->cbtext, aParity) == 0) {
        par = p;
        break;
      }
    }
    for(i = 0, p = COMSTOPBITS; i < NCOMSTOPBITS; i++, p++) {
      if (strcmp(p->cbtext, stopbits) == 0) {
        sb = p;
        break;
      }
    }
    retval = cbseropen(theHandle,
                       aComPort,
                       bd, db, par, sb, aTimeout,
                       rdCb, aCtx);
  }
  return retval;
} /* cbseropenstr() */

int cbseropenwstr(t_comport     *theHandle,
                  const wchar_t *aComPort,
                  const wchar_t *aSpeed,
                  const wchar_t *databits,
                  const wchar_t *aParity,
                  const wchar_t *stopbits,
                  int           aTimeout,
                  t_cbsercb     rdCb,
                  void          *aCtx) {
  char comPort[CBSTXTLEN+1], speed[CBSTXTLEN+1], db[CBSTXTLEN+1],
       parity[CBSTXTLEN+1], sb[CBSTXTLEN+1];
  int  retval;

  retval = 0;
  cbserinitcom(theHandle);
  if (aSpeed != NULL && databits != NULL &&
      aParity != NULL && stopbits != NULL) {
    wcstombs(comPort, aComPort, CBSTXTLEN); comPort[CBSTXTLEN] = '\0';
    wcstombs(speed,   aSpeed,   CBSTXTLEN); speed[CBSTXTLEN] = '\0';
    wcstombs(db,      databits, CBSTXTLEN); db[CBSTXTLEN] = '\0';
    wcstombs(parity,  aParity,  CBSTXTLEN); parity[CBSTXTLEN] = '\0';
    wcstombs(sb,      stopbits, CBSTXTLEN); sb[CBSTXTLEN] = '\0';
    retval = cbseropenstr(theHandle,
                          comPort,
                          speed, db, parity, sb, aTimeout,
                          rdCb, aCtx);
  }
  return retval;
} /* cbseropenwstr() */

int cbserwrite(const t_comport *aHandle, const char *aBuf, size_t aLen) {
  const char *ptr;
  int        retval;

  retval = 0;
  if (aHandle != NULL &&
      aBuf != NULL &&
      aHandle->hcomport != CBINVALIDCOMPORT) {
#ifdef FOR_MS_WINDOWS
    DWORD written, remaining;
    BOOL  ret;

    remaining = (DWORD) aLen;
    ret = TRUE;
    ptr = aBuf;
    while(ret && remaining > 0) {
      ret = WriteFile(aHandle->hcomport, ptr, remaining, &written, NULL);
      if (ret && written > 0) {
        ptr += written;
        remaining -= written;
      }
      else break;
    } /* while */
    if (ret && remaining <= 0) retval = 1;
#else
    int written, remaining;

    remaining = (int) aLen;
    ptr = aBuf;
    while(remaining > 0) {
      written = write(aHandle->hcomport, ptr, remaining);
      if (written > 0) {
        ptr += written;
        remaining -= written;
      }
      else break;
    }
    if (remaining <= 0) retval = 1;
#endif
  }
  return retval;
} /* cbserwrite() */

int cbserclose(t_comport *theHandle) {
  int retval;

  retval = 0;
  if (theHandle != NULL) {
    theHandle->goon = 0;
    if (theHandle->hthr != CBINVALIDTHREAD) {
      cbthr_join(theHandle->hthr);
    }
    if (theHandle->hcomport != CBINVALIDCOMPORT) {
      _closeport(theHandle->hcomport);
    }
    cbserinitcom(theHandle);
    retval = 1;
  } /* if (theHandle != NULL) */
  return retval;
} /* cbserclose() */

int cbserinfo(t_cbserinfo *theHandle, t_cbserinfocb cbsicb, void *aCtx) {
  int retval;

  retval = 0;
  cbserinitinfo(theHandle);
  if (theHandle != NULL && cbsicb != NULL) {
    theHandle->goon = 1;
    theHandle->cbsicb = cbsicb;
    theHandle->ctx = aCtx;
    retval = cbthr_create(&(theHandle->hthr), _cbsithrfunc, theHandle);
  }
  return retval;
} /* cbserinfo() */

int cbserinfo_end(t_cbserinfo *theHandle, short doStop) {
  int retval;

  retval = 0;
  if (theHandle != NULL) {
    if (doStop != 0) theHandle->goon = 0;
    if (theHandle->hthr != CBINVALIDTHREAD) {
      retval = cbthr_join(theHandle->hthr);
    }
    else retval = 1;
    cbserinitinfo(theHandle);
  }
  return retval;
} /* cbserinfo_stop() */

