#ifndef __CBTHR_H_
#define __CBTHR_H_

/** \file cbthr.h
  * \brief OS independend thread implementation.
  *
  * Uses posix threads for UNIX and Windows Threads for Win32.
  * Example of use
  * \code
  * #include "cbthr.h"
  * #include <stdio.h>
  *
  * static CBTHRFUNCTYPE _func(void *params)
  * {
  *   unsigned int thr_id;
  *   int          *iparam;
  *
  *   if (params != NULL)
  *   {
  *     iparam = (int *) params;
  *     printf("Thread param %d\n", *iparam);
  *   }
  *   CBTHREND;
  * }
  *
  * int main()
  * {
  *   t_cbthr thr_handle[4];
  *   int     thr_param[4];
  *   int     i, j;
  *
  *   for(i=0; i<4; i++) thr_param[i] = 10 + i;
  *   for(i=0; i<4; i++)
  *   {
  *     if (cbthr_create(&thr_handle[i], _func, &thr_param[i]) == 0)
  *       fprintf(stderr, "Thread create error\n");
  *   }
  *   for(j=0; j<i; j++)
  *     if (cbthr_join(thr_handle[j]) == 0)
  *       fprintf(stderr, "Thread join error\n");
  *   return 0;
  * }
  * \endcode
  * \author Daniel Stoinski, Nagler & Company
  * \version $Revision: 48 $
  */

/** \def CBTHRFUNCTYPE
  * \brief OS dependend return type of a thread function
  */

/** \typedef t_cbthr
  * \brief OS dependend type of the thread handle
  */

/** \typedef t_cbthrfunc
  * \brief OS dependend type of the thread function pointer
  */

/** \typedef t_cbthrcs
  * \brief OS dependend type for a critical section.
  *
  * Example of use:
  * \code
  * #include "cbthr.h"
  *
  * static t_cbthrcs cs;
  *
  * static CBTHRFUNCTYPE _func(void *params)
  * {
  *   ...
  *   cbthr_enter_cs(&cs);
  *   ...
  *   cbthr_leave_cs(&cs);
  * }
  * ...
  *
  * int main()
  * {
  *   cbthr_init_cs(&cs);
  *   ...
  * }
  * \endcode
  */

#ifndef W32_THREADS
# if (defined(_WIN32) || (defined(UNDER_CE)))
#  define W32_THREADS
# endif
#endif

#ifndef FOR_MS_WINDOWS
# if (defined(W32_THREADS) && ! defined(__WINE__))
#  define FOR_MS_WINDOWS
# endif
#endif

#ifdef W32_THREADS

# include <windows.h>
# define CBTHRFUNCTYPE DWORD __stdcall
# define CBTHREND { ExitThread(0); return 0; }

# ifdef __cplusplus
extern "C" {
# endif

typedef HANDLE           t_cbthr;
typedef CRITICAL_SECTION t_cbthrcs;
typedef DWORD (__stdcall *t_cbthrfunc)(void *);

# ifdef __cplusplus
}
# endif

#else

# include <pthread.h>
# define CBTHRFUNCTYPE void*
# define CBTHREND { pthread_exit(NULL); return NULL; }

# ifdef __cplusplus
extern "C" {
# endif

typedef pthread_t       t_cbthr;
typedef pthread_mutex_t t_cbthrcs;
typedef CBTHRFUNCTYPE (*t_cbthrfunc)(void *);

# ifdef __cplusplus
}
# endif

#endif /* W32_THREADS */

#ifdef __cplusplus
extern "C" {
#endif

extern const t_cbthr CBINVALIDTHREAD;

/** \brief Creates a new thread.
  *
  * \param theThread    out, the created thread handle.
  * \param aThrFunc     in, pointer to the thread function.
  * \param thrFuncParms in, parameter for the thread function.
  * \return  != 0 for OK, 0 for error.
  */
int cbthr_create(t_cbthr     *theThread,
                 t_cbthrfunc aThrFunc,
                 void        *thrFuncParms);

/** \brief Joins the thread.
  *
  * \param aThread handle of the thread to join.
  * \return != 0 for OK, 0 for error.
  */
int cbthr_join(t_cbthr aThread);

/** \brief Initializes the given critical section 
  *
  * An instance of a critical section variable should be defined
  * globally. This function is called once before subsequent calls
  * to NC_thread_enter_cs() and NC_thread_leave_cs().
  *
  * The function does nothing, if the given pointer is NULL.
  * \param aCritSec  pointer fo the critical section variable.
  * \return 0 on success, != 0 on error.
  * \sa cbthr_enter_cs()
  * \sa cbthr_leave_cs()
  * \sa t_cbthrcs
  */
int cbthr_init_cs(t_cbthrcs *aCritSec);

/** \brief Enters a critical section.
  * \param aCritSec  pointer to the critical section variable.
  * \return 0 on success, != 0 on error.
  */
int cbthr_enter_cs(t_cbthrcs *aCritSec);

/** \brief Leaves a critical section.
  *
  * Releases the critical section for other threads.
  * \param aCritSec  pointer to the critical section variable.
  * \return 0 on success, != 0 on error.
  */
int cbthr_leave_cs(t_cbthrcs *aCritSec);

/** \brief Releases resources occupied by the critical section.
  * \param aCritSec  pointer to the critical section to release
  * \return 0 on success, != 0 on error.
  */
int cbthr_delete_cs(t_cbthrcs *aCritSec);

# ifdef __cplusplus
}
# endif

#endif /* __CBTHR_H_ */

